# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..command import RobotoCommandSet
from .cancel import cancel_command
from .logs import get_logs_command
from .show import show_command
from .status import status_command

commands = [
    cancel_command,
    get_logs_command,
    show_command,
    status_command,
]

command_set = RobotoCommandSet(
    name="invocations",
    help=(
        "Get logs and status history from invoked actions, including the option to cancel them."
    ),
    commands=commands,
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .validators import (
    remove_non_noneable_init_args,
    validate_nonzero_gitpath_specs,
)

__all__ = [
    "remove_non_noneable_init_args",
    "validate_nonzero_gitpath_specs",
]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Optional, Union

import pydantic
from pydantic import ConfigDict

from roboto.sentinels import NotSet, NotSetType

from .record import CollectionResourceRef


class CreateCollectionRequest(pydantic.BaseModel):
    """Request payload to create a collection"""

    description: Optional[str] = None
    name: Optional[str] = None
    resources: Optional[list[CollectionResourceRef]] = None
    tags: Optional[list[str]] = None


class UpdateCollectionRequest(pydantic.BaseModel):
    """Request payload to update a collection"""

    add_resources: Union[list[CollectionResourceRef], NotSetType] = NotSet
    add_tags: Union[list[str], NotSetType] = NotSet
    description: Optional[Union[NotSetType, str]] = NotSet
    name: Optional[Union[NotSetType, str]] = NotSet
    remove_resources: Union[list[CollectionResourceRef], NotSetType] = NotSet
    remove_tags: Union[list[str], NotSetType] = NotSet

    model_config = ConfigDict(json_schema_extra=NotSetType.openapi_schema_modifier)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .api import (
    QueryContext,
    QueryRecord,
    QueryScheme,
    QueryStatus,
    QueryStorageContext,
    QueryStorageScheme,
    QueryTarget,
    SubmitRoboqlQueryRequest,
    SubmitStructuredQueryRequest,
    SubmitTermQueryRequest,
)
from .client import Query, QueryClient
from .conditions import (
    Comparator,
    Condition,
    ConditionGroup,
    ConditionOperator,
    ConditionType,
)
from .specification import (
    QuerySpecification,
    SortDirection,
)
from .visitor import BaseVisitor, ConditionVisitor

__all__ = (
    "BaseVisitor",
    "Comparator",
    "Condition",
    "ConditionGroup",
    "ConditionOperator",
    "ConditionType",
    "ConditionVisitor",
    "Query",
    "QueryClient",
    "QueryContext",
    "QueryRecord",
    "QueryScheme",
    "QuerySpecification",
    "QueryStorageContext",
    "QueryStorageScheme",
    "QueryStatus",
    "QueryTarget",
    "SortDirection",
    "SubmitStructuredQueryRequest",
    "SubmitRoboqlQueryRequest",
    "SubmitTermQueryRequest",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ._optional import (
    IMPORT_ERROR_MSG_TEMPLATE,
    import_optional_dependency,
)

__all__ = (
    "IMPORT_ERROR_MSG_TEMPLATE",
    "import_optional_dependency",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections
import typing

import pydantic
from pydantic import ConfigDict

from roboto.sentinels import NotSet, NotSetType
from roboto.updates import MetadataChangeset

from .record import DirectoryRecord, FileRecord


class DeleteFileRequest(pydantic.BaseModel):
    """Request payload to delete a file"""

    uri: str


class FileRecordRequest(pydantic.BaseModel):
    """Upsert a file record."""

    file_id: str
    tags: list[str] = pydantic.Field(default_factory=list)
    metadata: dict[str, typing.Any] = pydantic.Field(default_factory=dict)


class ImportFileRequest(pydantic.BaseModel):
    """Request payload to import a file"""

    dataset_id: str
    description: typing.Optional[str] = None
    relative_path: str
    size: int
    uri: str


class QueryFilesRequest(pydantic.BaseModel):
    """Request payload to query files"""

    filters: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    model_config = ConfigDict(extra="forbid")


class RenameFileRequest(pydantic.BaseModel):
    """Request payload to rename a file"""

    association_id: str
    new_path: str


class SignedUrlResponse(pydantic.BaseModel):
    """Response to a signed url request"""

    url: str


class UpdateFileRecordRequest(pydantic.BaseModel):
    """Request payload to update a file record"""

    description: typing.Optional[typing.Union[str, NotSetType]] = NotSet
    metadata_changeset: typing.Union[MetadataChangeset, NotSetType] = NotSet

    model_config = pydantic.ConfigDict(
        extra="forbid", json_schema_extra=NotSetType.openapi_schema_modifier
    )


class DirectoryContentsPage(pydantic.BaseModel):
    """Contents of a dataset directory page"""

    files: collections.abc.Sequence[FileRecord]
    directories: collections.abc.Sequence[DirectoryRecord]
    next_token: typing.Optional[str] = None



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import json
import typing

from mcap.decoder import (
    DecoderFactory as McapDecoderFactory,
)
from mcap.records import Schema
from mcap.well_known import MessageEncoding


class JsonDecoderFactory(McapDecoderFactory):
    """Provides functionality to an mcap.reader.McapReader to decode Json messages."""

    def decoder_for(
        self, message_encoding: str, schema: typing.Optional[Schema]
    ) -> typing.Optional[typing.Callable[[bytes], typing.Any]]:
        if message_encoding != MessageEncoding.JSON:
            return None

        def decoder(data: bytes):
            deserialized = data.decode()
            return json.loads(deserialized)

        return decoder



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import enum
import typing

import pydantic

from ...pydantic import (
    validate_nonzero_gitpath_specs,
)
from ...query import ConditionType
from .action_record import (
    ActionReference,
    ComputeRequirements,
    ContainerParameters,
)
from .invocation_record import (
    InvocationDataSource,
)


class TriggerForEachPrimitive(str, enum.Enum):
    """
    Trigger for each enum.
    """

    Dataset = "dataset"
    DatasetFile = "dataset_file"


class TriggerRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of an trigger.
    """

    trigger_id: str
    name: str
    org_id: str
    created: datetime.datetime
    created_by: str
    modified: datetime.datetime
    modified_by: str
    action: ActionReference
    required_inputs: list[str]
    service_user_id: str
    for_each: TriggerForEachPrimitive = TriggerForEachPrimitive.Dataset
    enabled: bool = True
    parameter_values: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    additional_inputs: typing.Optional[list[str]] = None
    compute_requirement_overrides: typing.Optional[ComputeRequirements] = None
    container_parameter_overrides: typing.Optional[ContainerParameters] = None
    condition: typing.Optional[ConditionType] = None
    timeout: typing.Optional[int] = None

    @pydantic.field_validator("required_inputs")
    def validate_required_inputs(cls, value: list[str]) -> list[str]:
        return validate_nonzero_gitpath_specs(value)

    @pydantic.field_validator("additional_inputs")
    def validate_additional_inputs(
        cls, value: typing.Optional[list[str]]
    ) -> typing.Optional[list[str]]:
        if value is None or len(value) == 0:
            return []

        return validate_nonzero_gitpath_specs(value)


class TriggerEvaluationCause(enum.Enum):
    """
    The cause of a TriggerEvaluationRecord is the reason why the trigger was selected for evaluation.
    """

    DatasetMetadataUpdate = "dataset_metadata_update"
    FileUpload = "file_upload"


class TriggerEvaluationStatus(enum.Enum):
    """
    When a trigger is selected for evaluation,
    a trigger evaluation record is created with a status of Pending.
    The evaluation can either run to completion (regardless of its outcome),
    in which case the status is Evaluated, or hit an unexpected exception,
    in which case the status is Failed.
    """

    Pending = "pending"
    Evaluated = "evaluated"
    Failed = "failed"


class TriggerEvaluationOutcome(enum.Enum):
    """
    The outcome of a TriggerEvaluationRecord is the result of the evaluation.
    A trigger can either invoke its associated action (one or many times) or be skipped.
    If skipped, a skip reason is provided.
    """

    InvokedAction = "invoked_action"
    Skipped = "skipped"


class TriggerEvaluationOutcomeReason(enum.Enum):
    """Context for why a trigger evaluation has its TriggerEvaluationOutcome"""

    AlreadyRun = "already_run"
    """
    This trigger has already run its associated action for this dataset and/or file.
    """

    ConditionNotMet = "condition_not_met"
    """
    The trigger's condition is not met.
    """

    NoMatchingFiles = "no_matching_files"
    """
    In the case of a dataset trigger,
    there is no subset of files that, combined, match ALL of the trigger's required inputs.

    In the case of a file trigger,
    there are no files that match ANY of the trigger's required inputs.
    """

    TriggerDisabled = "trigger_disabled"
    """
    The trigger is disabled.
    """


class TriggerEvaluationRecord(pydantic.BaseModel):
    """
    Record of a point-in-time evaluation of whether to invoke an action associated with a trigger for a data source.
    """

    trigger_evaluation_id: int  # Auto-generated by the database
    trigger_id: str
    data_source: InvocationDataSource
    evaluation_start: datetime.datetime
    evaluation_end: typing.Optional[datetime.datetime] = None
    status: TriggerEvaluationStatus
    status_detail: typing.Optional[str] = (
        None  # E.g., exception that caused the evaluation to fail
    )
    outcome: typing.Optional[TriggerEvaluationOutcome] = None
    outcome_reason: typing.Optional[TriggerEvaluationOutcomeReason] = None
    cause: typing.Optional[TriggerEvaluationCause] = None



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..command import RobotoCommandSet
from .changes import changes_command
from .create import create_command
from .delete import delete_command
from .list import list_command
from .show import show_command
from .update import update_command

command_set = RobotoCommandSet(
    name="collections",
    help="Curate collections of datasets and other data types.",
    commands=[
        changes_command,
        create_command,
        delete_command,
        list_command,
        show_command,
        update_command,
    ],
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import enum
import sys
import typing


class AnsiColor(str, enum.Enum):
    RED = "\033[0;31m"
    GREEN = "\033[0;32m"
    BLUE = "\033[0;34m"
    END = "\033[0m"


def print_error_and_exit(msg: typing.Union[str, list[str]]) -> typing.NoReturn:
    if isinstance(msg, list):
        for line in msg:
            print(f"{AnsiColor.RED}[error]{AnsiColor.END} {line}", file=sys.stderr)
    else:
        print(f"{AnsiColor.RED}[error]{AnsiColor.END} {msg}", file=sys.stderr)
    sys.exit(1)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import abc
import collections
import collections.abc
import typing

from .record import MessagePathRecord

KNOWN_PATH_MAPPING: dict[str, dict[str, tuple[str, ...]]] = {
    "header.stamp.sec": {
        "sec": (
            # ROS1
            # https://github.com/foxglove/mcap/blob/main/python/mcap-ros1-support/mcap_ros1/_vendor/genpy/rostime.py#L54
            "secs",
            # ROS2: a mapping isn't needed
            # https://github.com/foxglove/mcap/blob/main/python/mcap-ros2-support/mcap_ros2/_dynamic.py#L131
        ),
    },
    "header.stamp.nsec": {
        "nsec": (
            # ROS1
            # https://github.com/foxglove/mcap/blob/main/python/mcap-ros1-support/mcap_ros1/_vendor/genpy/rostime.py#L54
            "nsecs",
            # ROS2
            # https://github.com/foxglove/mcap/blob/main/python/mcap-ros2-support/mcap_ros2/_dynamic.py#L131
            "nanosec",
        )
    },
}


class AttrGetter(abc.ABC):
    """Abstract collection of utilities used to interact with values returned from MCAP decoders"""

    @staticmethod
    @abc.abstractmethod
    def get_attribute_names(value) -> collections.abc.Sequence[str]: ...

    @staticmethod
    @abc.abstractmethod
    def get_attribute(value, attribute) -> typing.Any: ...

    @staticmethod
    @abc.abstractmethod
    def has_sub_attributes(value) -> bool: ...


class ClassAttrGetter(AttrGetter):
    """Getter for ROS decoded data, which are yielded as dynamically created classes at runtime"""

    @staticmethod
    def get_attribute_names(value):
        return value.__slots__

    @staticmethod
    def get_attribute(value, attribute):
        return getattr(value, attribute)

    @staticmethod
    def has_sub_attributes(value):
        return hasattr(value, "__slots__")


class DictAttrGetter(AttrGetter):
    """Getter for JSON decoded data, which are yielded as dictionaries at runtime"""

    @staticmethod
    def get_attribute_names(value):
        return value.keys()

    @staticmethod
    def get_attribute(value, attribute):
        return value[attribute]

    @staticmethod
    def has_sub_attributes(value):
        return hasattr(value, "keys")


class DecodedMessage:
    """
    Facade for value returned from an MCAP Message decoder.

    A decoded message may be one a few different types:
      - A dictionary in the case the message data is encoded as JSON
      - A dynamically created, custom object type in the case the message data is encoded as ROS1 or ROS2 (CDR).
    """

    __message: typing.Union[dict, typing.Type]
    __message_paths: collections.abc.Sequence[MessagePathRecord]

    def __init__(
        self,
        msg: typing.Union[dict, typing.Type],
        message_paths: collections.abc.Sequence[MessagePathRecord],
    ):
        self.__message = msg
        self.__message_paths = message_paths

    def to_dict(self) -> dict:
        accumulator: dict[str, typing.Any] = {}

        getter: AttrGetter = ClassAttrGetter()

        if isinstance(self.__message, dict):
            # Message data was encoded as JSON
            getter = DictAttrGetter()

        # Assume a 1:1 relation between slots and message fields
        # Ref:
        #   - https://github.com/foxglove/mcap/blob/main/python/mcap-ros1-support/mcap_ros1/_vendor/genpy/message.py#L334-L338  # noqa E501
        #   - https://github.com/foxglove/mcap/blob/main/python/mcap-ros2-support/mcap_ros2/_dynamic.py#L258
        for attribute in getter.get_attribute_names(self.__message):
            if any(
                record.message_path.startswith(attribute)
                for record in self.__message_paths
            ):
                self.__unpack(
                    accumulator=accumulator,
                    obj=self.__message,
                    attr=attribute,
                    attr_path=attribute,
                    getter=getter,
                )

        return accumulator

    def __unpack(
        self,
        accumulator: dict[str, typing.Any],
        obj: typing.Any,
        attr: str,
        attr_path: str,
        getter: AttrGetter,
    ):
        value = getter.get_attribute(obj, attr)
        if not getter.has_sub_attributes(value):
            accumulator[attr] = value
        else:
            accumulator[attr] = {}
            for slot in getter.get_attribute_names(value):
                next_attr_path = ".".join([attr_path, slot])
                if any(
                    record.message_path.startswith(next_attr_path)
                    for record in self.__message_paths
                ):
                    self.__unpack(
                        accumulator=accumulator[attr],
                        obj=value,
                        attr=slot,
                        attr_path=next_attr_path,
                        getter=getter,
                    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json

from ...image_registry import ImageRegistry
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def list_images(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    image_registry = ImageRegistry(context.roboto_client)
    paginated_results = image_registry.list_images(
        repository_name=args.repository_name, org_id=args.org
    )
    while True:
        for image in paginated_results.items:
            print(json.dumps(image.model_dump(), indent=2))
        if paginated_results.next_token:
            paginated_results = image_registry.list_images(
                repository_name=args.repository_name,
                page_token=paginated_results.next_token,
                org_id=args.org,
            )
        else:
            break


def list_images_parser(parser: argparse.ArgumentParser) -> None:
    add_org_arg(parser)

    parser.add_argument(
        "--repository-name",
        dest="repository_name",
        action="store",
        help="Filter list of images by repository name.",
    )


ls_images_command = RobotoCommand(
    name="list",
    logic=list_images,
    setup_parser=list_images_parser,
    command_kwargs={"help": "List container images hosted in Roboto's image registry."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing
from typing import Any, Optional
import urllib.parse

from roboto.notifications import (
    NotificationChannel,
    NotificationType,
)

from ...http import RobotoClient, roboto_headers
from .operations import (
    CreateUserRequest,
    UpdateUserRequest,
)
from .record import UserRecord


class User:
    """
    Represents an individual who has access to the Roboto platform.
    """

    __record: UserRecord
    __roboto_client: RobotoClient

    @classmethod
    def create(
        cls,
        request: CreateUserRequest,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "User":
        """
        Creates a new user in Roboto. This API is only used by the Roboto platform itself as part of the signup
        process, any other caller will get an Unauthorized response from Roboto service.
        """
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.post("v1/users", data=request).to_record(UserRecord)
        return cls(record, roboto_client)

    @classmethod
    def for_self(cls, roboto_client: typing.Optional[RobotoClient] = None) -> "User":
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get("v1/users").to_record(UserRecord)
        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def from_id(
        cls, user_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ) -> "User":
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get(
            f"v1/users/id/{urllib.parse.quote_plus(user_id)}"
        ).to_record(UserRecord)
        return cls(record=record, roboto_client=roboto_client)

    def __init__(
        self, record: UserRecord, roboto_client: typing.Optional[RobotoClient] = None
    ):
        self.__record = record
        self.__roboto_client = RobotoClient.defaulted(roboto_client)

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def user_id(self) -> str:
        return self.__record.user_id

    @property
    def name(self) -> Optional[str]:
        return self.__record.name

    @property
    def record(self) -> UserRecord:
        return self.__record

    def delete(self) -> None:
        self.__roboto_client.delete(
            "v1/users", headers=roboto_headers(user_id=self.user_id)
        )

    def to_dict(self) -> dict[str, Any]:
        return self.__record.model_dump(mode="json")

    def update(
        self,
        name: Optional[str] = None,
        picture_url: Optional[str] = None,
        notification_channels_enabled: Optional[dict[NotificationChannel, bool]] = None,
        notification_types_enabled: Optional[dict[NotificationType, bool]] = None,
    ) -> "User":
        request = UpdateUserRequest(
            name=name,
            picture_url=picture_url,
            notification_channels_enabled=notification_channels_enabled,
            notification_types_enabled=notification_types_enabled,
        )
        self.__record = self.__roboto_client.put(
            "v1/users", headers=roboto_headers(user_id=self.user_id), data=request
        ).to_record(UserRecord)
        return self



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import enum
import json
import typing

import pydantic

from .action_record import (
    ComputeRequirements,
    ContainerParameters,
    ExecutorContainer,
)


class InvocationDataSourceType(enum.Enum):
    """Source of data for an action's input binding"""

    Dataset = "Dataset"


class InvocationDataSource(pydantic.BaseModel):
    """Abstracted data source that can be provided to an invocation"""

    data_source_type: InvocationDataSourceType
    # The "type" determines the meaning of "id":
    #   - if type is "Dataset," id is a dataset_id
    data_source_id: str


class ActionProvenance(pydantic.BaseModel):
    """Provenance information for an action"""

    name: str
    org_id: str
    # 2023-09-11 (GM): Optional for backwards compatibility; new invocations will always have a digest
    digest: typing.Optional[str] = None


class ExecutableProvenance(pydantic.BaseModel):
    """Provenance information for an action executable"""

    # Optional for backwards compatibility
    container_image_uri: typing.Optional[str] = None
    container_image_digest: typing.Optional[str] = None


class InvocationSource(enum.Enum):
    """Method by which an invocation was run"""

    Trigger = "Trigger"
    Manual = "Manual"


class SourceProvenance(pydantic.BaseModel):
    """Provenance information for an invocation source"""

    source_type: InvocationSource
    # The “type” determines the meaning of “id:”
    #   - if type is “Trigger,” id is a TriggerId;
    #   - if type is “Manual,” id is a UserId.
    source_id: str


class InvocationProvenance(pydantic.BaseModel):
    """Provenance information for an invocation"""

    action: ActionProvenance
    """The Action that was invoked."""

    executable: ExecutableProvenance
    """The underlying executable (e.g., Docker image) that was run."""

    source: SourceProvenance
    """The source of the invocation."""


class InvocationStatus(int, enum.Enum):
    """Invocation status enum"""

    Queued = 0
    Scheduled = 1
    Downloading = 2
    Processing = 3
    Uploading = 4
    Completed = 5
    # Failure status' and cancellation exists outside linear progression of invocation status
    Cancelled = 997
    Failed = 998
    Deadly = 999

    def __str__(self) -> str:
        return self.name

    @staticmethod
    def from_value(v: typing.Union[int, str]) -> "InvocationStatus":
        for value in InvocationStatus:
            if v in [value.value, value.name]:
                return value

        raise ValueError(f"Illegal Invocation status {v}")

    def can_transition_to(self, other: "InvocationStatus") -> bool:
        if self == other:
            return True

        if self in {
            InvocationStatus.Completed,
            InvocationStatus.Cancelled,
            InvocationStatus.Deadly,
        }:
            return False

        if self is InvocationStatus.Failed:
            if other in {InvocationStatus.Queued, InvocationStatus.Deadly}:
                return True
            return False

        if other in {InvocationStatus.Cancelled, InvocationStatus.Failed}:
            return True

        if other is InvocationStatus.Deadly:
            return self in {InvocationStatus.Queued, InvocationStatus.Failed}

        return other.value > self.value

    def is_running(self) -> bool:
        return self in {
            InvocationStatus.Downloading,
            InvocationStatus.Processing,
            InvocationStatus.Uploading,
        }

    def is_terminal(self) -> bool:
        return self in {
            InvocationStatus.Completed,
            InvocationStatus.Cancelled,
            InvocationStatus.Failed,
            InvocationStatus.Deadly,
        }

    def next(self) -> typing.Optional["InvocationStatus"]:
        if self.is_terminal():
            return None
        return InvocationStatus(self.value + 1)


class InvocationStatusRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of an invocation status.
    """

    status: InvocationStatus
    detail: typing.Optional[str] = None
    timestamp: datetime.datetime  # Persisted as ISO 8601 string in UTC

    def to_presentable_dict(self) -> dict[str, typing.Optional[str]]:
        return {
            "status": str(self.status),
            "timestamp": self.timestamp.isoformat(),
            "detail": self.detail,
        }


class LogsLocation(pydantic.BaseModel):
    """
    Invocation log storage location
    """

    bucket: str
    prefix: str


class InvocationRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of an invocation.
    """

    # When adding or removing fields, make sure to update __str__
    created: datetime.datetime  # Persisted as ISO 8601 string in UTC
    data_source: InvocationDataSource
    input_data: list[str]
    invocation_id: str  # Sort key
    idempotency_id: typing.Optional[str] = None
    compute_requirements: ComputeRequirements
    container_parameters: ContainerParameters
    last_heartbeat: typing.Optional[datetime.datetime] = None
    last_status: InvocationStatus
    org_id: str  # Partition key
    parameter_values: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    provenance: InvocationProvenance
    status: list[InvocationStatusRecord] = pydantic.Field(default_factory=list)
    duration: datetime.timedelta = pydantic.Field(default_factory=datetime.timedelta)
    timeout: int

    def __str__(self) -> str:
        return json.dumps(
            {
                "created": self.created.isoformat(),
                "data_source": self.data_source.model_dump(mode="json"),
                "input_data": self.input_data,
                "invocation_id": self.invocation_id,
                "idempotency_id": self.idempotency_id,
                "compute_requirements": self.compute_requirements.model_dump(
                    mode="json"
                ),
                "container_parameters": self.container_parameters.model_dump(
                    mode="json"
                ),
                "last_heartbeat": self.last_heartbeat,
                "last_status": self.last_status,
                "org_id": self.org_id,
                "parameter_values": self.parameter_values,
                "provenance": self.provenance.model_dump(mode="json"),
                "status": [
                    status_record.to_presentable_dict() for status_record in self.status
                ],
                "duration": str(self.duration),
                "timeout": self.timeout,
            },
            indent=2,
        )


class LogRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of a log record.
    """

    # If a log record is a partial log, this is a correlation ID for its parts
    # See documentation in the InvocationAwsDelegate in the method responsible for serving logs.
    partial_id: typing.Optional[str] = None
    log: str
    process: ExecutorContainer
    timestamp: datetime.datetime



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


COLLECTION_ID_HELP = "A unique ID used to reference a collection of Roboto resources."
COLLECTION_VERSION_HELP = (
    "An optional version for the provided collection ID."
    + " Allows caller to ensure that they're referencing an immutable representation of a given collection."
)
CONTENT_MODE_HELP = (
    "The type of content to return for a collection or set of collections. ``summary_only`` returns "
    + "only metadata, ``references`` will return each id in the collection, "
    + "and ``full`` will retrieve the full content of each resource in the collection."
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import email.message
import io
import json
import typing
import urllib.response

from .http_client import HttpResponse


class FakeHttpResponseFactory:
    """
    A factory for creating fake HTTP responses, for use with the roboto.http.HttpClient.

    Example:
        >>> import contextlib
        >>> import unittest.mock
        >>> from roboto.http import HttpClient, FakeHttpResponseFactory
        >>> mock_http_client = unittest.mock.create_autospec(HttpClient, instance=True)
        >>> with contextlib.ExitStack() as stack:
        ...     http_get_mock = stack.enter_context(
        ...         unittest.mock.patch.object(mock_http_client, "get")
        ...     )
        ...     http_get_mock.side_effect = FakeHttpResponseFactory(
        ...         "https://example.com",
        ...         {"foo": "bar"},
        ...         status_code=200,
        ...         headers={"Content-Type": "application/json"},
        ...     )

    """

    __headers: dict[str, str]
    __response_data: typing.Any
    __status_code: int
    __url: str

    def __init__(
        self,
        url: str = "https://iamverylazyanddonotseturls.com",
        response_data: typing.Any = "{}",
        status_code: int = 200,
        headers: typing.Optional[dict[str, str]] = None,
    ) -> None:
        self.__status_code = status_code
        self.__headers = headers or dict()
        self.__response_data = response_data
        self.__url = url

    def __call__(self, *args, **kwargs) -> HttpResponse:
        headers = email.message.Message()
        for k, v in self.__headers.items():
            headers.add_header(k, v)
        data = io.BytesIO(json.dumps(self.__response_data).encode())
        urllib_response = urllib.response.addinfourl(
            data, headers, self.__url, self.__status_code
        )
        return HttpResponse(urllib_response)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import collections.abc
import json
import pathlib
import typing

from ...domain import actions
from ...exceptions import RobotoConflictException
from ...updates import MetadataChangeset
from ..command import (
    KeyValuePairsAction,
    RobotoCommand,
)
from ..common_args import (
    ActionParameterArg,
    ActionTimeoutArg,
    add_action_reference_arg,
    add_compute_requirements_args,
    add_container_parameters_args,
    add_org_arg,
    parse_compute_requirements,
    parse_container_overrides,
)
from ..context import CLIContext
from ..terminal import print_error_and_exit
from ..validation import (
    pydantic_validation_handler,
)
from .action_config import ActionConfig


def determine_updates(
    action_config: ActionConfig, existing_action: actions.Action
) -> collections.abc.Mapping:
    """Create an updates dict matching Action::update kwargs from an ActionConfig and an existing Action."""

    updates: dict[str, typing.Any] = dict()
    if action_config.compute_requirements != existing_action.compute_requirements:
        updates["compute_requirements"] = action_config.compute_requirements

    if action_config.container_parameters != existing_action.container_parameters:
        updates["container_parameters"] = action_config.container_parameters

    if action_config.description != existing_action.record.description:
        updates["description"] = action_config.description

    if action_config.short_description != existing_action.record.short_description:
        updates["short_description"] = action_config.short_description

    if action_config.timeout != existing_action.record.timeout:
        updates["timeout"] = action_config.timeout

    metadata_changeset_builder = MetadataChangeset.Builder()
    existing_keys = existing_action.record.metadata.keys()
    new_keys = action_config.metadata.keys()

    # Remove keys that are in the existing metadata but not in the new metadata
    for key in existing_keys - new_keys:
        metadata_changeset_builder.remove_field(key)

    # Add/overwrite keys that are in the new metadata
    for key in new_keys:
        metadata_changeset_builder.put_field(key, action_config.metadata[key])

    existing_tags = set(existing_action.record.tags)
    new_tags = set(action_config.tags)

    # Remove tags that are in the existing tags but not in the new tags
    for tag in existing_tags - new_tags:
        metadata_changeset_builder.remove_tag(tag)

    # Add tags that are in the new tags
    for tag in new_tags:
        metadata_changeset_builder.put_tag(tag)

    metadata_changeset = metadata_changeset_builder.build()
    if not metadata_changeset.is_empty():
        updates["metadata_changeset"] = metadata_changeset

    parameter_changeset_builder = actions.ActionParameterChangeset.Builder()
    existing_params = {param.name for param in existing_action.record.parameters}
    new_params = {param.name for param in action_config.parameters}

    # Remove parameters that are in the existing parameters but not in the new parameters
    for param_name in existing_params - new_params:
        parameter_changeset_builder.remove_parameter(param_name)

    # Add/overwrite parameters that are in the new parameters
    for param in action_config.parameters:
        parameter_changeset_builder.put_parameter(param)

    parameter_changeset = parameter_changeset_builder.build()
    if not parameter_changeset.is_empty():
        updates["parameter_changeset"] = parameter_changeset

    if action_config.image_uri != existing_action.uri:
        updates["uri"] = action_config.image_uri

    if action_config.inherits:
        if not action_config.inherits.owner:
            action_config.inherits.owner = existing_action.record.org_id

    if action_config.inherits != existing_action.record.inherits:
        updates["inherits"] = action_config.inherits

    return updates


def create(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    config: typing.Optional[ActionConfig] = None
    if args.action_config_file:
        if not args.action_config_file.exists():
            print_error_and_exit(
                f"Action config file '{args.action_config_file}' does not exist."
            )
        with pydantic_validation_handler("Action config file"):
            config = ActionConfig.model_validate_json(
                args.action_config_file.read_text()
            )

    default_compute_reqs_from_file = config.compute_requirements if config else None
    compute_requirements = parse_compute_requirements(
        args, defaults=default_compute_reqs_from_file
    )

    default_container_parameters_from_file = (
        config.container_parameters if config else None
    )
    container_parameters = parse_container_overrides(
        args, defaults=default_container_parameters_from_file
    )

    metadata = config.metadata if config else dict()
    if args.metadata:
        metadata.update(args.metadata)

    tags = config.tags if config else list()
    if args.tag:
        tags.extend(args.tag)

    parameters = config.parameters if config else list()
    if args.parameter:
        cli_param_names = {param.name for param in args.parameter}
        # replace any existing parameters with the same name
        parameters = [
            param for param in parameters if param.name not in cli_param_names
        ]
        parameters.extend(args.parameter)

    if not args.name and not config:
        print_error_and_exit(
            "Action name is required. Please specify either the `--name` CLI argument or the `name` property in your "
            "Action config file.",
        )

    config_params = {
        "name": args.name,
        "description": args.description,
        "inherits": args.inherits_from,
        "compute_requirements": compute_requirements,
        "container_parameters": container_parameters,
        "metadata": metadata,
        "parameters": args.parameter,
        "tags": tags,
        "image_uri": args.image,
        "short_description": args.short_description,
        "timeout": args.timeout,
    }
    config_params = {k: v for k, v in config_params.items() if v is not None}

    with pydantic_validation_handler("Action configuration"):
        config_with_overrides = (
            ActionConfig.model_validate(config_params)
            if config is None
            else config.model_copy(update=config_params)
        )

    image_uri = None
    if config_with_overrides.docker_config:
        # Build docker image and push it to Roboto's Docker register
        print_error_and_exit(
            [
                "Support for building and pushing Docker images as part of Action creation is forthcoming.",
                "For now, please use the `docker` CLI to build and tag your image, and use `roboto images push` "
                "to push it to Roboto's Docker registry.",
                "Finally, use either the `--image` CLI argument or the `image_uri` property in your Action config file "
                "to associate your Action with the Docker image you just pushed.",
            ]
        )
    else:
        image_uri = config_with_overrides.image_uri

    try:
        action = actions.Action.create(
            name=config_with_overrides.name,
            parameters=config_with_overrides.parameters,
            uri=image_uri,
            inherits=config_with_overrides.inherits,
            description=config_with_overrides.description,
            compute_requirements=config_with_overrides.compute_requirements,
            container_parameters=config_with_overrides.container_parameters,
            metadata=config_with_overrides.metadata,
            tags=config_with_overrides.tags,
            short_description=config_with_overrides.short_description,
            timeout=config_with_overrides.timeout,
            caller_org_id=args.org,
            roboto_client=context.roboto_client,
        )

        print(f"Successfully created action '{action.name}'. Record: ")
        print(json.dumps(action.to_dict(), indent=2))
    except RobotoConflictException:
        action = actions.Action.from_name(
            name=config_with_overrides.name,
            owner_org_id=args.org,
            roboto_client=context.roboto_client,
        )

        if not args.yes:
            print(
                f"Action '{action.name}' already exists. Do you want to update it? [y/n]"
            )
            choice = input().lower()
            if choice not in ["y", "yes"]:
                return

        updates = determine_updates(config_with_overrides, action)

        if updates:
            action.update(**updates)
            print(f"Successfully updated action '{action.name}'. Record: ")
            print(json.dumps(action.to_dict(), indent=2))
        else:
            print(f"Action '{action.name}' is up-to-date. Record: ")
            print(json.dumps(action.to_dict(), indent=2))


def create_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--from-file",
        type=pathlib.Path,
        action="store",
        dest="action_config_file",
        help=(
            "Path to a file containing Action configuration. "
            "Other CLI arguments specified will override the values in the config file. "
        ),
    )
    parser.add_argument(
        "--name",
        required=False,
        action="store",
        help=(
            "Name of the action. Not modifiable after creation. "
            "An action is considered unique by its (name, docker_image_name, docker_image_tag) tuple."
        ),
    )

    parser.add_argument(
        "--description",
        required=False,
        action="store",
        help="Optional description of action. Modifiable after creation.",
    )

    parser.add_argument(
        "--short_description",
        required=False,
        action="store",
        help="Optional description of action. Modifiable after creation.",
    )

    add_action_reference_arg(
        parser=parser,
        arg_name="inherits_from",
        arg_help=(
            "Partially or fully qualified reference to action from which to inherit configuration. "
            "Inheriting from another action is mutually exclusive with specifying a container image (``--image``), "
            "entrypoint (``--entrypoint``), command (``--command``), working directory (``--workdir``), "
            "env vars (``--env``), or parameter(s) (``--parameter``). "
        ),
        positional=False,
        required=False,
    )
    parser.add_argument(
        "--image",
        required=False,
        action="store",
        dest="image",
        help="Associate a Docker image with this action. Modifiable after creation.",
    )
    parser.add_argument(
        "--parameter",
        required=False,
        metavar=ActionParameterArg.METAVAR,
        nargs="*",
        action=ActionParameterArg,
        help=(
            "Zero or more parameters (space-separated) accepted by this action. "
            "``name`` is the only required attribute. "
            "``default`` values, if provided, are JSON parsed. "
            "This argument can be specified multiple times. "
            "Parameters can be modified after creation. "
            "Argument values must be wrapped in quotes. E.g.: "
            "``--put-parameter 'name=my_param|required=true|description=My description of my_param'``"
        ),
    )
    parser.add_argument(
        "--metadata",
        required=False,
        metavar="KEY=VALUE",
        nargs="*",
        action=KeyValuePairsAction,
        help=(
            "Zero or more ``<key>=<value>`` format key/value pairs which represent action metadata. "
            "``value`` is parsed as JSON. "
            "Metadata can be modified after creation."
        ),
    )
    parser.add_argument(
        "--tag",
        required=False,
        type=str,
        nargs="*",
        help="One or more tags to annotate this action. Modifiable after creation.",
        action="extend",
    )
    parser.add_argument(
        "--timeout",
        required=False,
        action=ActionTimeoutArg,
        help="Optional timeout for an action in minutes, defaults to 30 minutes or 12 hours depending on tier.",
    )
    add_org_arg(parser=parser)

    add_compute_requirements_args(parser)
    add_container_parameters_args(parser)

    parser.add_argument(
        "--yes",
        required=False,
        help="If action with same name already exists, update the existing action without prompting for confirmation.",
        action="store_true",
    )


create_command = RobotoCommand(
    name="create",
    logic=create,
    setup_parser=create_parser,
    command_kwargs={"help": "Create a new action."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import datetime
import enum
import typing

import pydantic

from ...association import Association


class RepresentationStorageFormat(enum.Enum):
    """
    Supported representation format types
    """

    MCAP = "mcap"


class RepresentationRecord(pydantic.BaseModel):
    """
    Pointer to data extracted from a Topic--potentially a specific slice of the data,
    from within a MessagePath--processed for representation in a particular modality.

    Most MessagePaths within a Topic will point to the same Representation
    (e.g., an MCAP file with all of the Topic data).
    Some MessagePaths, however, may have more than one Representation,
    such as both an MCAP and a preview GIF or statically rendered, line-chart preview.
    """

    association: Association
    """
    Identifier and entity type with which this Representation is associated. E.g., a file, a database.
    """

    representation_id: str
    storage_format: RepresentationStorageFormat
    topic_id: str
    version: int


class CanonicalDataType(enum.Enum):
    """
    Well-known (and simplified) data types, normalized across frameworks/technologies.
    Primarily used for UI purposes to determine if a Panel can render a particular MessagePath.

    References:

        - ROS 1 field types: http://wiki.ros.org/msg
        - ROS 2 field types: https://docs.ros.org/en/iron/Concepts/Basic/About-Interfaces.html#field-types
        - uORB: there is no list of primitive field types, but they appear
          to be the same as ROS: https://docs.px4.io/main/en/middleware/uorb.html#adding-a-new-topic

    Example mappings:

        - ``float32`` -> ``DataType.Number``
        - ``uint8[]`` -> ``DataType.Array``
        - ``uint8[]`` -> ``DataType.Image``
        - ``geometry_msgs/Pose`` -> ``DataType.Object``
        - ``std_msgs/Header`` -> ``DataType.Object``
        - ``string`` -> ``DataType.String``
        - ``char`` -> ``DataType.String``
        - ``bool`` -> ``DataType.Boolean``
        - ``byte`` -> ``DataType.Byte``

    """

    Array = "array"  # A sequence of values.
    Boolean = "boolean"
    Byte = "byte"
    Image = "image"  # Special purpose type for data that can be rendered as an image.
    Number = "number"
    Object = "object"  # A struct with attributes.
    String = "string"
    Unknown = "unknown"  # This is a fallback and should be used sparingly.

    # Special purpose types for data that represents geographic points
    LatDegFloat = "latdegfloat"  # e.g. 47.6749387 (used in ULog ver_data_format >= 2)
    LonDegFloat = "londegfloat"  # e.g. 9.1445274 (used in ULog ver_data_format >= 2)
    LatDegInt = "latdegint"  # e.g. 317534036 (used in ULog ver_data_format < 2)
    LonDegInt = "londegint"  # e.g. 1199146398 (used in ULog ver_data_format < 2)


class MessagePathStatistic(enum.Enum):
    """Statistics computed by Roboto in our standard ingestion actions."""

    Count = "count"
    Max = "max"
    Mean = "mean"
    Median = "median"
    Min = "min"


class MessagePathRecord(pydantic.BaseModel):
    """
    Path to a typed attribute within individual datum records contained within a Topic.
    """

    canonical_data_type: CanonicalDataType
    """Normalized data type, used primarily internally by the Roboto Platform."""

    created: datetime.datetime
    created_by: str

    data_type: str
    """
    'Native'/framework-specific data type of the attribute at this path.
    E.g. "float32", "uint8[]", "geometry_msgs/Pose", "string".
    """

    message_path: str
    """
    Dot-delimited path to the attribute within the datum record.
    """

    metadata: collections.abc.Mapping[str, typing.Any] = pydantic.Field(
        default_factory=dict,
    )
    """
    Key-value pairs to associate with this metadata for discovery and search, e.g.
    `{ 'min': '0.71', 'max': '1.77 }`
    """

    modified: datetime.datetime
    modified_by: str

    representations: collections.abc.MutableSequence[RepresentationRecord] = (
        pydantic.Field(default_factory=list)
    )
    """
    Zero to many Representations of this MessagePath.
    """

    topic_id: str
    message_path_id: str


class TopicRecord(pydantic.BaseModel):
    """
    Collection of timestamped data that share a common name and association (e.g., file).

    Data from the same file within the same named topic/channel are considered part of the same Topic.
    Data from different files or with differently named topics/channels belong within unique Topics.

    Source files that have been chunked by time or size but from the same point-in-time data collection will
    produce multiple Topics for the same "logical topic" (~= name, schema) within those chunks.
    """

    association: Association
    """
    Identifier and entity type with which this Topic is associated. E.g., a file, a dataset.
    """

    created: datetime.datetime

    created_by: str

    default_representation: typing.Optional[RepresentationRecord] = None
    """
    Default Representation for this Topic.
    Assume that if a MessagePath is not more specifically associated with a Representation,
    it should use this one.
    """

    end_time: typing.Optional[int] = None
    """
    Timestamp of oldest message in topic, in nanoseconds since epoch (assumed Unix epoch).
    """

    message_count: typing.Optional[int] = None

    message_paths: collections.abc.MutableSequence[MessagePathRecord] = pydantic.Field(
        default_factory=list
    )
    """
    Zero to many MessagePathRecords associated with this TopicSource.
    """

    metadata: collections.abc.Mapping[str, typing.Any] = pydantic.Field(
        default_factory=dict
    )
    """
    Arbitrary metadata.
    """

    modified: datetime.datetime

    modified_by: str

    org_id: str

    schema_checksum: typing.Optional[str] = None
    """
    Checksum of topic schema.
    May be None if topic does not have a known/named schema.
    """

    schema_name: typing.Optional[str] = None
    """
    Type of messages in topic. E.g., "sensor_msgs/PointCloud2".
    May be None if topic does not have a known/named schema.
    """

    schema_version: int = 0
    """
    Dynamically-determined at write-time.
    A more human-friendly way by which to compare schemas than schema_checksum.
    Schemas with the same (name, checksum) ought to have the same version.
    It communicates, "this is the Nth version of a topic with this name that has been uploaded to Roboto".
    """

    start_time: typing.Optional[int] = None
    """
    Timestamp of earliest message in topic, in nanoseconds since epoch (assumed Unix epoch).
    """

    topic_id: str
    topic_name: str



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Optional

import pydantic

from roboto.notifications import (
    NotificationChannel,
    NotificationType,
)


class CreateUserRequest(pydantic.BaseModel):
    """
    Request payload to create a new user
    """

    user_id: str
    name: Optional[str] = None
    is_service_user: bool = False
    is_system_user: bool = False
    picture_url: Optional[str] = None
    default_notification_channels: Optional[list[NotificationChannel]] = [
        NotificationChannel.Email
    ]
    default_notification_types: Optional[list[NotificationType]] = [
        NotificationType.CommentMention
    ]


class UpdateUserRequest(pydantic.BaseModel):
    """
    Request payload to update a user
    """

    name: Optional[str] = None
    picture_url: Optional[str] = None
    notification_channels_enabled: Optional[dict[NotificationChannel, bool]] = None
    notification_types_enabled: Optional[dict[NotificationType, bool]] = None



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .args import (
    ExistingPathlibPath,
    JsonFileOrStrType,
    KeyValuePairsAction,
)
from .roboto_command import (
    RobotoCommand,
    RobotoCommandSet,
)

__all__ = [
    "ExistingPathlibPath",
    "JsonFileOrStrType",
    "KeyValuePairsAction",
    "RobotoCommand",
    "RobotoCommandSet",
]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime

import pydantic

from .validator import (
    LifecycleStatusValidator,
    NotificationChannel,
    NotificationType,
    ReadStatus,
)


class NotificationRecord(pydantic.BaseModel, LifecycleStatusValidator):
    """
    A wire-transmissible representation of a notification.
    """

    notification_id: str
    org_id: str
    user_id: str
    notifier_id: str
    notification_type: NotificationType
    channels: list[NotificationChannel] = pydantic.Field(default_factory=list)
    lifecycle_status: dict[NotificationChannel, str] = pydantic.Field(
        default_factory=dict
    )
    read_status: ReadStatus = ReadStatus.Unread
    created: datetime.datetime
    modified: datetime.datetime



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import json
import typing
import urllib.error
import urllib.parse
import urllib.request
import urllib.response

import pydantic

from .retry import (
    RetryWaitFn,
    default_retry_wait_ms,
)

Model = typing.TypeVar("Model")


class HttpRequest:
    url: str
    method: str
    headers: dict
    retry_wait: RetryWaitFn
    data: typing.Any = None
    idempotent: bool = False

    def __init__(
        self,
        url: str,
        method: str = "GET",
        headers: typing.Optional[dict[str, str]] = None,
        data: typing.Any = None,
        retry_wait: typing.Optional[RetryWaitFn] = None,
        idempotent: bool = False,
    ):
        self.url = url
        self.method = method
        self.headers = headers if headers is not None else {}
        self.data = data
        self.retry_wait = (
            retry_wait if retry_wait is not None else default_retry_wait_ms
        )
        self.idempotent = idempotent

        if isinstance(self.data, pydantic.BaseModel) or isinstance(self.data, dict):
            self.headers["Content-Type"] = "application/json"

    def __repr__(self) -> str:
        return (
            f"HttpRequest("
            f"url={self.url}, "
            f"method={self.method}, "
            f"headers={self.headers}, "
            f"data={self.data}, "
            f"idempotent={self.idempotent}"
            ")"
        )

    @property
    def body(self) -> typing.Optional[bytes]:
        if self.data is None:
            return None

        if isinstance(self.data, bytes):
            return self.data

        if isinstance(self.data, str):
            return self.data.encode("utf-8")

        if isinstance(self.data, pydantic.BaseModel):
            return self.data.model_dump_json(exclude_unset=True).encode("utf-8")

        return json.dumps(self.data).encode("utf-8")

    @property
    def hostname(self) -> str:
        parsed_url = urllib.parse.urlparse(self.url)
        return (
            parsed_url.hostname
            if parsed_url.hostname is not None
            else parsed_url.netloc
        )

    def append_headers(self, headers: dict[str, str]) -> None:
        self.headers.update(headers)


HttpRequestDecorator = typing.Callable[[HttpRequest], HttpRequest]


class BatchRequest(pydantic.BaseModel, typing.Generic[Model]):
    """Batched HTTP requests"""

    requests: list[Model]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import typing

import pydantic

from ...association import Association


class EventRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of an event.
    """

    associations: list[Association] = pydantic.Field(default_factory=list)
    """
    Datasets, files, topics and message paths which this event pertains to.
    """

    created: datetime.datetime
    """
    Date/time when this event was created.
    """

    created_by: str = pydantic.Field(description="The user who registered this device.")
    """
    The user who created this event.
    """

    description: typing.Optional[str] = None
    """
    An optional human-readable description of the event.
    """

    end_time: int
    """
    The end time of the event, in nanoseconds since epoch (assumed Unix epoch). This can be equal to start_time if
    the event is discrete, but can never be less than start_time.
    """

    event_id: str
    """
    A globally unique ID used to reference an event.
    """

    metadata: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    """
    Key-value pairs to associate with this event for discovery and search.
    """

    modified: datetime.datetime
    """
    Date/time when this device record was last modified.
    """

    modified_by: str
    """
    The user who last modified this device record.
    """

    name: str
    """
    A brief human-readable name for the event. Many events can have the same name. "Takeoff", "Crash", "CPU Spike",
    "Bad Image Quality", and "Unexpected Left" are a few potential examples.
    """

    org_id: str = pydantic.Field(description="The org to which this device belongs.")
    """
    The org to which this device belongs.
    """

    start_time: int
    """
    The start time of the event, in nanoseconds since epoch (assumed Unix epoch).
    """

    tags: list[str] = pydantic.Field(default_factory=list)
    """
    Tags to associate with this event for discovery and search.
    """

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, EventRecord):
            return NotImplemented

        return (
            sorted(
                self.associations,
                key=lambda a: (a.association_type.value, a.association_id),
            )
            == sorted(
                other.associations,
                key=lambda a: (a.association_type.value, a.association_id),
            )
            and self.created == other.created
            and self.created_by == other.created_by
            and self.description == other.description
            and self.end_time == other.end_time
            and self.event_id == other.event_id
            and self.metadata == other.metadata
            and self.modified == other.modified
            and self.modified_by == other.modified_by
            and self.name == other.name
            and self.org_id == other.org_id
            and self.start_time == other.start_time
            and set(self.tags) == set(other.tags)
        )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json

from ...domain.orgs import (
    Org,
    OrgInvite,
    OrgRoleName,
)
from ...domain.orgs.s3_integration import (
    S3IntegrationService,
)
from ..command import (
    RobotoCommand,
    RobotoCommandSet,
)
from ..common_args import get_defaulted_org_id
from ..context import CLIContext

GENERIC_ORG_ONLY_SETUP_PARSER_DEFAULT_HELP = "perform some action"


def generic_org_add_argument(
    parser, action_to_perform: str = GENERIC_ORG_ONLY_SETUP_PARSER_DEFAULT_HELP
):
    parser.add_argument(
        "--org",
        type=str,
        help=f"The org_id of the org for which to {action_to_perform}. "
        + "This parameter is only required if the caller is a member of more than one org, "
        + "otherwise this action will be performed implicitly on their single org.",
    )


def generic_org_only_setup_parser(
    action_to_perform: str = GENERIC_ORG_ONLY_SETUP_PARSER_DEFAULT_HELP,
):
    def setup(parser):
        generic_org_add_argument(parser=parser, action_to_perform=action_to_perform)

    return setup


def create(args, context: CLIContext, parser: argparse.ArgumentParser):
    org = Org.create(
        name=args.name,
        bind_email_domain=args.bind_email_domain,
        roboto_client=context.roboto_client,
    )
    print(json.dumps(org.to_dict(), indent=2))


def create_setup_parser(parser):
    parser.add_argument(
        "--name", type=str, required=True, help="A human readable name for this org"
    )
    parser.add_argument(
        "--bind-email-domain",
        action="store_true",
        help="Automatically add new users with your email domain to this org",
    )


def delete(args, context: CLIContext, parser: argparse.ArgumentParser):
    org = Org.from_id(org_id=args.org, roboto_client=context.roboto_client)

    if not args.ignore_prompt:
        print("Are you absolutely sure you want to delete your org? [y/n]: ")

        choice = input().lower()
        if choice not in ["y", "yes"]:
            return

    org.delete()
    print("Successfully deleted!")


def delete_setup_parser(parser):
    parser.add_argument(
        "--org",
        type=str,
        required=True,
        help="The org_id for the org you're about to delete.",
    )

    parser.add_argument(
        "--ignore-prompt",
        action="store_true",
        help="Ignore the prompt which asks you to confirm that you'd like to delete your org.",
    )


def show(args, context: CLIContext, parser: argparse.ArgumentParser):
    record = Org.from_id(org_id=args.org, roboto_client=context.roboto_client)
    print(json.dumps(record.to_dict(), indent=2))


def show_setup_parser(parser):
    parser.add_argument(
        "--org",
        type=str,
        help="The org_id for the org you want to see.",
    )


def list_org_members(args, context: CLIContext, parser: argparse.ArgumentParser):
    org_users = Org.from_id(args.org).users()
    for org_user in org_users:
        # We'll need to rethink filtering out service users once
        # they are more directly customer-facing
        if org_user.user.is_service_user:
            continue

        print(
            org_user.model_dump_json(
                indent=2,
                exclude={
                    "org",
                },
            )
        )


def list_org_members_setup_parser(parser):
    parser.add_argument(
        "--org",
        type=str,
        help="The org_id for the org you want to see.",
    )


def remove_user(args, context: CLIContext, parser: argparse.ArgumentParser):
    org = Org.from_id(org_id=args.org, roboto_client=context.roboto_client)
    org.remove_user(user_id=args.user)
    print("Successfully removed!")


def remove_user_setup_parser(parser):
    parser.add_argument(
        "--user",
        type=str,
        required=True,
        help="The user_id of the user to remove.",
    )
    parser.add_argument(
        "--org",
        type=str,
        help="The org_id of the org to remove a user from. "
        + "Required only if the caller is a member of more than one org.",
    )


def invite_user(args, context: CLIContext, parser: argparse.ArgumentParser):
    OrgInvite.create(
        invited_user_id=args.user,
        org_id=args.org,
        roboto_client=context.roboto_client,
    )
    print("Invite sent!")


def invite_user_setup_parser(parser):
    parser.add_argument(
        "--user",
        type=str,
        required=True,
        help="The user_id of the user to invite.",
    )
    parser.add_argument(
        "--org",
        help="The org_id of the org to invite a user to. "
        + "Required only if the caller is a member of more than one org.",
    )


def list_invites(args, context: CLIContext, parser: argparse.ArgumentParser):
    invites = OrgInvite.for_org(org_id=args.org, roboto_client=context.roboto_client)
    for invite in invites:
        print(json.dumps(invite.to_dict(), indent=2))


def list_invites_setup_parser(parser):
    parser.add_argument(
        "--org",
        help="The org_id of the org to view invites for. "
        + "Required only if the caller is a member of more than one org.",
    )


def add_role(args, context: CLIContext, parser: argparse.ArgumentParser):
    Org.from_id(org_id=args.org, roboto_client=context.roboto_client).add_role_for_user(
        user_id=args.user, role=args.role
    )
    print("Added!")


def add_role_setup_parser(parser):
    parser.add_argument(
        "--user",
        type=str,
        required=True,
        help="The user_id of the user to add a role for.",
    )
    parser.add_argument(
        "--role",
        type=OrgRoleName,
        choices=[OrgRoleName.admin, OrgRoleName.owner],
        help="The role to grant to the specified user",
    )
    parser.add_argument(
        "--org",
        type=str,
        help="The org_id of the org for which to add permissions for the specified user. "
        + "Required only if the caller is a member of more than one org.",
    )


def remove_role(args, context: CLIContext, parser: argparse.ArgumentParser):
    Org.from_id(
        org_id=args.org, roboto_client=context.roboto_client
    ).remove_role_from_user(user_id=args.user, role=args.role)
    print("Removed!")


def remove_role_setup_parser(parser):
    parser.add_argument(
        "--user",
        type=str,
        required=True,
        help="The user_id of the user to add a role for. "
        + "This can be your own user_id if you would like to step down as an admin or owner.",
    )
    parser.add_argument(
        "--role",
        type=OrgRoleName,
        choices=[OrgRoleName.admin, OrgRoleName.owner],
        help="The role to remove for the specified user",
    )
    parser.add_argument(
        "--org",
        type=str,
        help="The org_id of the org for which to remove permissions for the specified user. "
        + "Required only if the caller is a member of more than one org.",
    )


def bind_email_domain(args, context: CLIContext, parser: argparse.ArgumentParser):
    Org.from_id(org_id=args.org, roboto_client=context.roboto_client).bind_email_domain(
        args.email_domain
    )
    print(f"Successfully bound domain {args.email_domain}")


def bind_email_domain_setup_parser(parser):
    generic_org_add_argument(parser=parser, action_to_perform="bind an email domain")
    parser.add_argument(
        "--email-domain",
        type=str,
        required=True,
        help="The email domain to bind to an org.",
    )


def unbind_email_domain(args, context: CLIContext, parser: argparse.ArgumentParser):
    Org.from_id(
        org_id=args.org, roboto_client=context.roboto_client
    ).unbind_email_domain(args.email_domain)
    print(f"Successfully unbound domain {args.email_domain}")


def unbind_email_domain_setup_parser(parser):
    generic_org_add_argument(parser=parser, action_to_perform="unbind an email domain")
    parser.add_argument(
        "--email-domain",
        type=str,
        required=True,
        help="The email domain to unbind from an org.",
    )


def list_email_domains(args, context: CLIContext, parser: argparse.ArgumentParser):
    for domain in Org.from_id(
        org_id=args.org, roboto_client=context.roboto_client
    ).email_domains():
        print(domain)


def register_bucket(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    service = S3IntegrationService(context.roboto_client)
    org_id = get_defaulted_org_id(args.org)

    service.register_bucket(
        org_id=org_id,
        bucket_name=args.bucket_name,
        account_id=args.aws_account,
    )

    print(
        "Bucket registered successfully! It will be used to store files for all new datasets."
    )


def register_bucket_setup_parser(parser: argparse.ArgumentParser) -> None:
    generic_org_add_argument(parser, "register an S3 bring-your-own-bucket")
    parser.add_argument(
        "--bucket-name",
        type=str,
        required=True,
        help="The base name of the S3 bucket you want to register, e.g. 'my-company-roboto-byob'",
    )
    parser.add_argument(
        "--aws-account",
        type=str,
        required=True,
        help="The AWS account to register the bucket in. This command will only work if you also have the calling "
        + "environment configured with AWS credentials for this account that the underlying boto3 client can use. "
        + "For more information on configuring AWS credentials, see "
        + "https://docs.aws.amazon.com/sdkref/latest/guide/standardized-credentials.html.",
    )


create_command = RobotoCommand(
    name="create",
    logic=create,
    setup_parser=create_setup_parser,
    command_kwargs={"help": "Creates a new organization"},
)


delete_command = RobotoCommand(
    name="delete",
    logic=delete,
    setup_parser=delete_setup_parser,
    command_kwargs={"help": "Deletes an existing organization"},
)


show_command = RobotoCommand(
    name="show",
    logic=show,
    setup_parser=show_setup_parser,
    command_kwargs={"help": "Gets metadata for a single organization"},
)


list_org_members_command = RobotoCommand(
    name="members",
    logic=list_org_members,
    setup_parser=list_org_members_setup_parser,
    command_kwargs={"help": "Lists the members of an organization"},
)

remove_user_command = RobotoCommand(
    name="remove-user",
    logic=remove_user,
    setup_parser=remove_user_setup_parser,
    command_kwargs={"help": "Removes a user from an organization"},
)

invite_command = RobotoCommand(
    name="invite-user",
    logic=invite_user,
    setup_parser=invite_user_setup_parser,
    command_kwargs={"help": "Invites a user to join an org."},
)

list_invites_command = RobotoCommand(
    name="list-invites",
    logic=list_invites,
    setup_parser=list_invites_setup_parser,
    command_kwargs={"help": "Lists the current pending invites for a specified org"},
)

bind_email_domain_command = RobotoCommand(
    name="bind-email-domain",
    logic=bind_email_domain,
    setup_parser=bind_email_domain_setup_parser,
    command_kwargs={
        "help": "Binds an email domain to this org, so all new users whose "
        + "emails are part of that domain will automatically be added."
    },
)

unbind_email_domain_command = RobotoCommand(
    name="unbind-email-domain",
    logic=unbind_email_domain,
    setup_parser=unbind_email_domain_setup_parser,
    command_kwargs={
        "help": "Unbinds an email domain from this org, so all new users whose "
        + "emails are part of that domain will no longer be automatically added."
    },
)

list_email_domains_command = RobotoCommand(
    name="list-email-domains",
    logic=list_email_domains,
    setup_parser=generic_org_only_setup_parser(
        action_to_perform="list bound email domains"
    ),
    command_kwargs={
        "help": "Lists the email domains associated with a particular org."
    },
)

add_role_command = RobotoCommand(
    name="add-role",
    logic=add_role,
    setup_parser=add_role_setup_parser,
    command_kwargs={"help": "Promotes a user to a more permissive org access level."},
)

remove_role_command = RobotoCommand(
    name="remove-role",
    logic=remove_role,
    setup_parser=remove_role_setup_parser,
    command_kwargs={"help": "Demotes a user to a less permissive org access level."},
)

register_bucket_command = RobotoCommand(
    name="register-bucket",
    logic=register_bucket,
    setup_parser=register_bucket_setup_parser,
    command_kwargs={
        "help": "Registers an S3 bring-your-own-bucket with Roboto. Only available to premium tier orgs."
    },
)

commands = [
    add_role_command,
    bind_email_domain_command,
    delete_command,
    invite_command,
    list_email_domains_command,
    list_invites_command,
    list_org_members_command,
    register_bucket_command,
    remove_role_command,
    remove_user_command,
    show_command,
    unbind_email_domain_command,
]

command_set = RobotoCommandSet(
    name="orgs",
    help="View details of organizations that you are in, including administration options if applicable.",
    commands=commands,
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .file import File
from .operations import (
    DeleteFileRequest,
    DirectoryContentsPage,
    FileRecordRequest,
    ImportFileRequest,
    QueryFilesRequest,
    RenameFileRequest,
    SignedUrlResponse,
    UpdateFileRecordRequest,
)
from .record import (
    CredentialProvider,
    DirectoryRecord,
    FileRecord,
    FileStatus,
    FileStorageType,
    FileTag,
    IngestionStatus,
    S3Credentials,
)

__all__ = (
    "CredentialProvider",
    "DeleteFileRequest",
    "File",
    "DirectoryContentsPage",
    "FileRecord",
    "FileRecordRequest",
    "FileStatus",
    "FileStorageType",
    "FileTag",
    "ImportFileRequest",
    "IngestionStatus",
    "DirectoryRecord",
    "NoopProgressMonitor",
    "NoopProgressMonitorFactory",
    "ProgressMonitor",
    "ProgressMonitorFactory",
    "QueryFilesRequest",
    "RenameFileRequest",
    "S3Credentials",
    "SignedUrlResponse",
    "UpdateFileRecordRequest",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import enum
import platform
import typing

import pydantic

from ..version import (
    __version__ as roboto_version,
)


class RobotoTool(str, enum.Enum):
    """Tool used to access Roboto"""

    Cli = "cli"
    Sdk = "sdk"
    UploadAgent = "upload-agent"
    Website = "website"


class RobotoRequester(pydantic.BaseModel):
    """
    Details about the entity making a request to Roboto. These are embedded in a header in order to see what tool
    versions / operating systems are making requests, and to aid debugging.
    """

    schema_version: typing.Literal["v1"] = "v1"
    """Roboto Requester payload schema version, used to ensure backward compatibility"""

    platform: typing.Optional[str] = None
    """The environment in which a request is being made, i.e. the user agent (for browser requests) or the results
    of platform.platform (for SDK requests)"""

    roboto_tool: typing.Optional[typing.Union[RobotoTool, str]] = None
    """If a request is being made from a Roboto vended tool, the name of the tool"""

    roboto_tool_version: typing.Optional[str] = None
    """If a request is being made from a Roboto vended tool, the version of the tool"""

    roboto_tool_details: typing.Optional[str] = None
    """If a request is being made from a Roboto vended tool, free text pertinent details about the tool"""

    @classmethod
    def for_tool(cls, tool: RobotoTool) -> "RobotoRequester":
        """
        Called to intelligently populate a :class:`~RobotoRequester` for a request made from a named Roboto tool
        using the Python SDK.
        """

        return RobotoRequester(
            schema_version="v1",
            platform=platform.platform(),
            roboto_tool=tool.value,
            roboto_tool_version=roboto_version,
            roboto_tool_details=None,
        )


ROBOTO_REQUESTER_HEADER = "X-Roboto-Requester"
"""A JSON serialized :class:`~RobotoRequester` representing the entity making a request to Roboto."""



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing

import pydantic

# Python 3.8/3.9 compatible import of TypeGuard
try:
    from typing import TypeGuard
except ImportError:
    try:
        from typing_extensions import TypeGuard
    except ImportError:
        pass

T = typing.TypeVar("T")
PydanticModel = typing.TypeVar("PydanticModel", bound=pydantic.BaseModel)


class Null:
    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return "null"


null = Null()


class NotSetType(pydantic.BaseModel):
    """
    A sentinel value for fields (e.g. kwargs) that are not set, useful when 'None' is a meaningful value.

    It extends pydantic.BaseModel so that it integrates nicely with pydantic's schema generation/json serialization.
    """

    def __init__(self, **data):
        if len(data) > 0:
            raise ValueError(f"Unexpected data: {data} passed into NotSetType")

        super().__init__(**data)

    @staticmethod
    def openapi_schema_modifier(
        schema: dict[str, typing.Any], model: typing.Any
    ) -> None:
        """
        Replace NotSetType with null in anyOf lists.

        To use in pydantic.BaseModel, add the following to your model:
        >>> class Config:
        ...     schema_extra = NotSetType.openapi_schema_modifier


        See https://docs.pydantic.dev/1.10/usage/schema/#schema-customization
        """
        for prop in schema.get("properties", {}).values():
            if "anyOf" in prop:
                any_of = prop.pop("anyOf")

                without_notset = []
                for sub_prop in any_of:
                    if isinstance(sub_prop, dict) and "NotSetType" in sub_prop.get(
                        "$ref", ""
                    ):
                        if "default" in prop and prop["default"] == {}:
                            del prop["default"]
                        without_notset.append({"type": "null"})
                    else:
                        without_notset.append(sub_prop)

                if without_notset:
                    if len(without_notset) == 1:
                        prop.update(without_notset[0])
                    else:
                        prop["anyOf"] = without_notset

    def __bool__(self) -> bool:
        return False


def is_not_set(value: typing.Any) -> TypeGuard[NotSetType]:
    return isinstance(value, NotSetType)


def is_set(value: typing.Union[T, NotSetType]) -> TypeGuard[T]:
    return not isinstance(value, NotSetType)


def value_or_not_set(value: typing.Optional[T]) -> typing.Union[T, NotSetType]:
    if value is None:
        return NotSet
    else:
        return value


def value_or_default_if_unset(value: typing.Union[T, NotSetType], default: T) -> T:
    if isinstance(value, NotSetType):
        return default
    else:
        return value


def remove_not_set(value: PydanticModel) -> PydanticModel:
    """
    This function round-trips a pydantic model through initialization while ensuring that NotSetType values are not
    explicitly set. This allows value.model_dump(exclude_unset=True) to be used to generate a correct payload.
    """
    set_args: dict[typing.Any, typing.Any] = {
        k: v for k, v in dict(value).items() if is_set(v)
    }
    return value.model_validate(set_args)


def is_any_set(value: PydanticModel) -> bool:
    """
    Checks whether a Pydantic model has at least one field that is explicitly set.
    """
    for _, v in dict(value).items():
        if is_set(v):
            return True

    return False


NotSet = NotSetType()



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import abc
from typing import Any

from ..command import (
    RobotoCommand,
    RobotoCommandSet,
)
from ..context import CLIContext


class RobotoCLIExtension(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def get_name(cls) -> str:
        raise NotImplementedError("get_name")

    @classmethod
    @abc.abstractmethod
    def get_command_sets(self) -> list[RobotoCommandSet]:
        raise NotImplementedError("get_command_sets")

    @classmethod
    @abc.abstractmethod
    def get_commands(cls) -> list[tuple[str, RobotoCommand]]:
        raise NotImplementedError("get_commands_extensions")

    @classmethod
    @abc.abstractmethod
    def get_context(cls, base_context: CLIContext) -> Any:
        raise NotImplementedError("get_context")


def apply_roboto_cli_command_extensions(
    base_command_sets: list[RobotoCommandSet],
) -> list[RobotoCommandSet]:
    command_sets_by_name: dict[str, RobotoCommandSet] = {}
    for command_set in base_command_sets:
        command_sets_by_name[command_set.name] = command_set

    for subclass in RobotoCLIExtension.__subclasses__():
        for command_set in subclass.get_command_sets():
            if command_set.name in command_sets_by_name.keys():
                raise ValueError(
                    f"Attempting to add already defined command set '{command_set.name}'"
                )
            command_sets_by_name[command_set.name] = command_set

    # Run this as a 2nd for loop, so we can define all command sets before we attempt to extend them.
    for subclass in RobotoCLIExtension.__subclasses__():
        for command_set_name, command in subclass.get_commands():
            if command_set_name not in command_sets_by_name.keys():
                raise ValueError(
                    f"Attempting to add command '{command.name}' to non-existant "
                    + "command set '{command_set_name}'"
                )

            command_sets_by_name[command_set_name].commands.append(command)

    return list(map(lambda item: item[1], command_sets_by_name.items()))


def apply_roboto_cli_context_extensions(base_context: CLIContext):
    for subclass in RobotoCLIExtension.__subclasses__():
        base_context.extensions[subclass.get_name()] = subclass.get_context(
            base_context
        )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...domain.collections import (
    Collection,
    CollectionResourceRef,
    CollectionResourceType,
)
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def create(args, context: CLIContext, parser: argparse.ArgumentParser):
    resources: list[CollectionResourceRef] = []

    if args.dataset_id is not None:
        resources.extend(
            [
                CollectionResourceRef(
                    resource_type=CollectionResourceType.Dataset, resource_id=dataset_id
                )
                for dataset_id in args.dataset_id
            ]
        )

    if args.file_id is not None:
        resources.extend(
            [
                CollectionResourceRef(
                    resource_type=CollectionResourceType.File, resource_id=file_id
                )
                for file_id in args.file_id
            ]
        )

    collection = Collection.create(
        name=args.name,
        description=args.description,
        tags=args.tag,
        resources=None if len(resources) == 0 else resources,
        caller_org_id=args.org,
        roboto_client=context.roboto_client,
    )
    print(collection.record.model_dump_json(indent=2))


def create_setup_parser(parser):
    parser.add_argument(
        "--name",
        type=str,
        help="A human-readable name for this collection. "
        + "Does not need to be unique.",
    )
    parser.add_argument(
        "--description", type=str, help="Information about what's in this collection"
    )
    parser.add_argument(
        "--dataset-id",
        nargs="*",
        help="Initial datasets to add to this collection.",
        action="extend",
    )
    parser.add_argument(
        "--file-id",
        nargs="*",
        help="Initial files to add to this collection.",
        action="extend",
    )
    parser.add_argument(
        "--tag",
        type=str,
        nargs="*",
        help="Tags that make it easier to discover this collection.",
        action="extend",
    )
    add_org_arg(parser)


create_command = RobotoCommand(
    name="create",
    logic=create,
    setup_parser=create_setup_parser,
    command_kwargs={"help": "Creates a new collection."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
from datetime import datetime
import pathlib
import typing

from ...association import Association
from ...compat import import_optional_dependency
from ...http import RobotoClient
from ...time import Time
from .record import (
    CanonicalDataType,
    MessagePathRecord,
    MessagePathStatistic,
)
from .topic_data_service import TopicDataService

if typing.TYPE_CHECKING:
    import pandas  # pants: no-infer-dep

PreComputedStat: typing.TypeAlias = typing.Optional[typing.Union[int, float]]


class MessagePath:
    """
    The set of data created by mapping over :py:class:`~roboto.domain.topics.Topic` data,
    picking out just those values from each datum at a given attribute path.
    """

    DELIMITER: typing.ClassVar = "."

    __record: MessagePathRecord
    __roboto_client: RobotoClient
    __topic_data_service: TopicDataService

    @staticmethod
    def parents(path: str) -> list[str]:
        """
        Given message path in dot notation, return list of its parent paths (also in dot notation).

        Example:
            >>> path = "pose.pose.position.x"
            >>> MessagePath.parents(path)
            ['pose.pose.position', 'pose.pose', 'pose']
        """
        parent_parts = MessagePath.parts(path)[:-1]
        return [
            MessagePath.DELIMITER.join(parent_parts[:i])
            for i in range(len(parent_parts), 0, -1)
        ]

    @staticmethod
    def parts(path: str) -> list[str]:
        """
        Split message path in dot notation into its constituent path parts.

        Example:
            >>> path = "pose.pose.position.x"
            >>> MessagePath.parts(path)
            ['pose', 'pose', 'position', 'x']
        """
        return path.split(MessagePath.DELIMITER)

    @classmethod
    def from_id(
        cls,
        message_path_id: str,
        roboto_client: typing.Optional[RobotoClient] = None,
        topic_data_service: typing.Optional[TopicDataService] = None,
    ) -> "MessagePath":
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get(
            f"v1/topics/message-path/id/{message_path_id}"
        ).to_record(MessagePathRecord)
        return cls(record, roboto_client, topic_data_service)

    def __init__(
        self,
        record: MessagePathRecord,
        roboto_client: typing.Optional[RobotoClient] = None,
        topic_data_service: typing.Optional[TopicDataService] = None,
    ):
        self.__record = record
        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__topic_data_service = topic_data_service or TopicDataService(
            self.__roboto_client
        )

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, MessagePath):
            return NotImplemented

        return self.__record == other.__record

    @property
    def canonical_data_type(self) -> CanonicalDataType:
        """Canonical Roboto data type corresponding to the native data type."""

        return self.__record.canonical_data_type

    @property
    def count(self) -> PreComputedStat:
        return self.__get_statistic(MessagePathStatistic.Count)

    @property
    def created(self) -> datetime:
        return self.__record.created

    @property
    def created_by(self) -> str:
        return self.__record.created_by

    @property
    def data_type(self) -> str:
        """Native data type for this message path, e.g. 'float32'"""

        return self.__record.data_type

    @property
    def max(self) -> PreComputedStat:
        return self.__get_statistic(MessagePathStatistic.Max)

    @property
    def mean(self) -> PreComputedStat:
        return self.__get_statistic(MessagePathStatistic.Mean)

    @property
    def median(self) -> PreComputedStat:
        return self.__get_statistic(MessagePathStatistic.Median)

    @property
    def message_path_id(self) -> str:
        return self.__record.message_path_id

    @property
    def metadata(self) -> dict[str, typing.Any]:
        return dict(self.__record.metadata)

    @property
    def min(self) -> PreComputedStat:
        return self.__get_statistic(MessagePathStatistic.Min)

    @property
    def modified(self) -> datetime:
        return self.__record.modified

    @property
    def modified_by(self) -> str:
        return self.__record.modified_by

    @property
    def path(self) -> str:
        return self.__record.message_path

    @property
    def record(self) -> MessagePathRecord:
        return self.__record

    @property
    def topic_id(self) -> str:
        return self.__record.topic_id

    def get_data(
        self,
        start_time: typing.Optional[Time] = None,
        end_time: typing.Optional[Time] = None,
        cache_dir: typing.Union[str, pathlib.Path, None] = None,
    ) -> collections.abc.Generator[dict[str, typing.Any], None, None]:
        """
        Return a projection of topic data: the set of data created by mapping over data collected within a topic,
        picking out just those values at this message path.

        If ``start_time`` or ``end_time`` are defined,
        they should either be integers that represent nanoseconds since UNIX epoch,
        or convertible to such by :py:func:`~roboto.time.to_epoch_nanoseconds`.
        Either or both may be omitted.
        ``start_time`` is inclusive, while ``end_time`` is exclusive.

        If ``cache_dir`` is defined, topic data will be downloaded to this location if necessary.
        If not provided, ``cache_dir`` defaults to
        :py:attr:`~roboto.domain.topics.topic_data_service.TopicDataService.DEFAULT_CACHE_DIR`.

        For each example below, assume the following is a sample datum record
        that can be found in this MessagePath's associated topic:

        ::

            {
                "angular_velocity": {
                    "x": <uint32>,
                    "y": <uint32>,
                    "z": <uint32>
                },
                "orientation": {
                    "x": <uint32>,
                    "y": <uint32>,
                    "z": <uint32>,
                    "w": <uint32>
                }
            }

        Examples:
            Print all data to stdout.

            >>> topic = Topic.from_name_and_association(...)
            >>> angular_velocity_x = topic.get_message_path("angular_velocity.x")
            >>> for record in angular_velocity_x.get_data():
            >>>      print(record)

            Collect data into a dataframe. Requires installing the ``roboto[analytics]`` extra.

            >>> import math
            >>> import pandas as pd
            >>> topic = Topic.from_name_and_association(...)
            >>> angular_velocity_x = topic.get_message_path("angular_velocity.x")
            >>> df = angular_velocity_x.get_data_as_df()
            >>> assert math.isclose(angular_velocity_x.mean, df[angular_velocity_x.path].mean())
        """

        yield from self.__topic_data_service.get_data(
            topic_id=self.__record.topic_id,
            message_paths_include=[self.__record.message_path],
            start_time=start_time,
            end_time=end_time,
            cache_dir_override=cache_dir,
        )

    def get_data_as_df(
        self,
        start_time: typing.Optional[Time] = None,
        end_time: typing.Optional[Time] = None,
        cache_dir: typing.Union[str, pathlib.Path, None] = None,
    ) -> pandas.DataFrame:
        """
        Return this message path's underlying data as a pandas DataFrame.

        Requires installing this package using the ``roboto[analytics]`` extra.

        See :py:meth:`~roboto.domain.topics.message_path.MessagePath.get_data` for more information on the parameters.
        """
        pandas = import_optional_dependency("pandas", "analytics")

        df = pandas.json_normalize(
            data=list(
                self.get_data(
                    start_time=start_time,
                    end_time=end_time,
                    cache_dir=cache_dir,
                )
            )
        )
        return df.set_index(TopicDataService.LOG_TIME_ATTR_NAME)

    def to_association(self) -> Association:
        return Association.msgpath(self.message_path_id)

    def __get_statistic(self, stat: MessagePathStatistic) -> PreComputedStat:
        return self.__record.metadata.get(stat.value)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import typing

from ...http import RobotoClient
from .org_records import OrgInviteRecord


class OrgInvite:
    """
    An invitation to join an organization from one user to another
    """

    __record: OrgInviteRecord
    __roboto_client: RobotoClient

    @classmethod
    def create(
        cls,
        invited_user_id: str,
        org_id: str,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "OrgInvite":
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.post(
            f"v1/orgs/id/{org_id}/users/id/{invited_user_id}/invites"
        ).to_record(OrgInviteRecord)
        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def from_id(
        cls, invite_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ):
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get(f"v1/orgs/invites/id/{invite_id}").to_record(
            OrgInviteRecord
        )
        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def for_org(
        cls, org_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ) -> collections.abc.Collection["OrgInvite"]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        records = roboto_client.get(f"v1/orgs/id/{org_id}/invites").to_record_list(
            OrgInviteRecord
        )
        return [cls(record=record, roboto_client=roboto_client) for record in records]

    def __init__(
        self,
        record: OrgInviteRecord,
        roboto_client: typing.Optional[RobotoClient] = None,
    ):
        self.__record = record
        self.__roboto_client = RobotoClient.defaulted(roboto_client)

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def invite_id(self) -> str:
        return self.__record.invite_id

    @property
    def invited_user_id(self) -> str:
        return self.__record.user_id

    @property
    def invited_by_user_id(self) -> str:
        return self.__record.invited_by.user_id

    @property
    def org_id(self) -> str:
        return self.__record.org.org_id

    def accept(self) -> None:
        self.__roboto_client.post(f"v1/orgs/invites/id/{self.invite_id}/accept")

    def decline(self) -> None:
        self.__roboto_client.post(f"v1/orgs/invites/id/{self.invite_id}/decline")

    def to_dict(self) -> dict[str, typing.Any]:
        return self.__record.model_dump(mode="json")



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import os
import pathlib
import sys
from typing import Optional

from packaging.version import Version
import pydantic

from ..http import HttpClient
from ..time import utcnow
from ..version import __version__
from .terminal import AnsiColor


class CLIState(pydantic.BaseModel):
    last_checked_version: Optional[datetime.datetime] = None
    last_version: str = "0.0.0"
    out_of_date: bool = True


def check_last_update():
    roboto_tmp_dir = pathlib.Path.home() / ".roboto" / "tmp"
    roboto_tmp_dir.mkdir(parents=True, exist_ok=True)
    cli_state_file = roboto_tmp_dir / "cli_state.json"

    last_version = None

    state = CLIState(last_checked_version=None)
    if cli_state_file.is_file():
        state = CLIState.model_validate_json(cli_state_file.read_text())
        last_version = state.last_version

    if (
        state.last_checked_version is None
        or __version__ != last_version
        or state.out_of_date is None
        or state.out_of_date is True
        or (utcnow() - datetime.timedelta(hours=1)) > state.last_checked_version
    ):
        http = HttpClient()

        releases = http.get(url="https://pypi.org/pypi/roboto/json").to_dict(
            json_path=["releases"]
        )
        versions = list(releases.keys())
        versions.sort(key=Version)
        latest = versions[-1]

        state.last_checked_version = utcnow()
        state.last_version = __version__
        state.out_of_date = __version__ != latest

        cli_state_file.write_text(state.model_dump_json())

        suppress_message = (
            os.getenv("ROBOTO_CLI_SUPPRESS_UPGRADE_PROMPT", "false").lower() != "false"
        )

        if state.out_of_date and not suppress_message:
            notice = f"{AnsiColor.BLUE}[notice]{AnsiColor.END}"
            print(
                f"\n{notice} A new release of roboto is available: "
                + f"{AnsiColor.RED + __version__ + AnsiColor.END} -> {AnsiColor.GREEN + latest + AnsiColor.END}\n"
                + f"{notice} To update, follow Upgrade CLI instructions at "
                + "https://github.com/roboto-ai/roboto-python-sdk/blob/main/README.md",
                file=sys.stderr,
            )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Any, Optional

from roboto.exceptions import (
    RobotoHttpExceptionParse,
)
from roboto.http import HttpClient

from .http_resources import (
    UpdateNotificationRequest,
)
from .record import NotificationChannel


class NotificationsClient:
    __http_client: HttpClient

    def __init__(self, http_client: HttpClient):
        super().__init__()
        self.__http_client = http_client

    def get_notifications(self) -> dict[str, Any]:
        url = self.__http_client.url("v1/notifications")

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url)

            return response.to_dict(json_path=["data"])

    def update_notification(
        self,
        notification_id: str,
        org_id: Optional[str] = None,
        read_status: Optional[str] = None,
        lifecycle_status: Optional[dict[NotificationChannel, str]] = None,
    ) -> Any:
        url = self.__http_client.url(f"v1/notifications/{notification_id}")

        body: dict[str, Any] = {
            "notification_id": notification_id,
        }

        request_headers: Optional[dict[str, str]] = None

        if org_id is not None:
            request_headers = {
                "X-Roboto-Org-Id": org_id,
            }

        if read_status is not None:
            body["read_status"] = read_status

        if lifecycle_status is not None:
            body["lifecycle_status"] = lifecycle_status

        with RobotoHttpExceptionParse():
            response = self.__http_client.put(
                url=url,
                data=body,
                headers=request_headers,
            )

            return response.to_dict(json_path=["data"])

    def batch_update_notifications(
        self, updates: list[UpdateNotificationRequest]
    ) -> dict[str, Any]:
        url = self.__http_client.url("v1/notifications/batch")

        dict_updates = []

        for update in updates:
            model_dict = update.model_dump()
            dict_updates.append(model_dict)

        with RobotoHttpExceptionParse():
            response = self.__http_client.put(
                url=url,
                data={
                    "requests": dict_updates,
                },
            )

            return response.to_dict(json_path=["data"])

    def delete_notification(self, notification_id: str) -> None:
        url = self.__http_client.url(f"v1/notifications/{notification_id}")

        with RobotoHttpExceptionParse():
            self.__http_client.delete(
                url=url,
            )

            return None



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
from datetime import datetime
import pathlib
import typing
import urllib.parse
import urllib.request

from ...association import Association
from ...compat import import_optional_dependency
from ...http import RobotoClient
from ...logging import default_logger
from ...sentinels import (
    NotSet,
    NotSetType,
    is_set,
    remove_not_set,
)
from ...time import Time
from ...updates import (
    MetadataChangeset,
    TaglessMetadataChangeset,
)
from .message_path import MessagePath
from .operations import (
    AddMessagePathRepresentationRequest,
    AddMessagePathRequest,
    CreateTopicRequest,
    MessagePathChangeset,
    SetDefaultRepresentationRequest,
    UpdateMessagePathRequest,
    UpdateTopicRequest,
)
from .record import (
    CanonicalDataType,
    MessagePathRecord,
    RepresentationRecord,
    RepresentationStorageFormat,
    TopicRecord,
)
from .topic_data_service import TopicDataService

if typing.TYPE_CHECKING:
    import pandas  # pants: no-infer-dep

logger = default_logger()


class Topic:
    """
    A sequence of structured time-series data linked to a source file.
    Each topic follows a defined schema, where the message paths represent the
    individual fields or signals within that schema.
    """

    __record: TopicRecord
    __roboto_client: RobotoClient
    __topic_data_service: TopicDataService

    @classmethod
    def create(
        cls,
        file_id: str,
        topic_name: str,
        end_time: typing.Optional[int] = None,
        message_count: typing.Optional[int] = None,
        metadata: typing.Optional[collections.abc.Mapping[str, typing.Any]] = None,
        schema_checksum: typing.Optional[str] = None,
        schema_name: typing.Optional[str] = None,
        start_time: typing.Optional[int] = None,
        message_paths: typing.Optional[
            collections.abc.Sequence[AddMessagePathRequest]
        ] = (None),
        caller_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Topic":
        roboto_client = RobotoClient.defaulted(roboto_client)
        request = CreateTopicRequest(
            association=Association.file(file_id),
            end_time=end_time,
            message_count=message_count,
            message_paths=message_paths,
            metadata=metadata,
            schema_checksum=schema_checksum,
            schema_name=schema_name,
            start_time=start_time,
            topic_name=topic_name,
        )
        response = roboto_client.post(
            "v1/topics",
            data=request,
            caller_org_id=caller_org_id,
        )
        record = response.to_record(TopicRecord)
        return cls(record, roboto_client)

    @classmethod
    def from_id(
        cls,
        topic_id: str,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Topic":
        roboto_client = RobotoClient.defaulted(roboto_client)

        response = roboto_client.get(
            f"v1/topics/id/{topic_id}",
        )
        record = response.to_record(TopicRecord)
        return cls(record, roboto_client)

    @classmethod
    def from_name_and_file(
        cls,
        topic_name: str,
        file_id: str,
        owner_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Topic":
        roboto_client = RobotoClient.defaulted(roboto_client)
        quoted_topic_name = urllib.parse.quote_plus(topic_name)
        encoded_association = Association.file(file_id).url_encode()

        response = roboto_client.get(
            f"v1/topics/association/{encoded_association}/name/{quoted_topic_name}",
            owner_org_id=owner_org_id,
        )
        record = response.to_record(TopicRecord)
        return cls(record, roboto_client)

    @classmethod
    def get_by_dataset(
        cls, dataset_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ):
        """
        List all topics associated with files of a dataset. If multiple files have topics with the same name (i.e.
        if a dataset has chunked files with the same schema), they'll be returned as separate topic objects.
        """
        roboto_client = RobotoClient.defaulted(roboto_client)
        page_token: typing.Optional[str] = None

        while True:
            query_params: dict[str, typing.Any] = {}
            if page_token:
                query_params["page_token"] = str(page_token)

            paginated_results = roboto_client.get(
                f"v1/datasets/{dataset_id}/topics",
                query=query_params,
            ).to_paginated_list(TopicRecord)

            for record in paginated_results.items:
                yield cls(record=record, roboto_client=roboto_client)
            if paginated_results.next_token:
                page_token = paginated_results.next_token
            else:
                break

    @classmethod
    def get_by_file(
        cls,
        file_id: str,
        owner_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> collections.abc.Generator["Topic", None, None]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        encoded_association = Association.file(file_id).url_encode()

        page_token: typing.Optional[str] = None
        while True:
            response = roboto_client.get(
                f"v1/topics/association/{encoded_association}",
                owner_org_id=owner_org_id,
                query={"page_token": page_token} if page_token else None,
            )
            paginated_results = response.to_paginated_list(TopicRecord)
            for topic_record in paginated_results.items:
                yield cls(topic_record, roboto_client)
            if paginated_results.next_token:
                page_token = paginated_results.next_token
            else:
                break

    def __init__(
        self,
        record: TopicRecord,
        roboto_client: typing.Optional[RobotoClient] = None,
        topic_data_service: typing.Optional[TopicDataService] = None,
    ):
        self.__record = record
        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__topic_data_service = topic_data_service or TopicDataService(
            self.__roboto_client
        )

    def __eq__(self, other: typing.Any):
        if not isinstance(other, Topic):
            return NotImplemented

        return self.__record == other.__record

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def association(self) -> Association:
        return self.__record.association

    @property
    def created(self) -> datetime:
        return self.__record.created

    @property
    def created_by(self) -> str:
        return self.__record.created_by

    @property
    def dataset_id(self) -> typing.Optional[str]:
        if self.association.is_dataset:
            return self.association.association_id

        if self.association.is_file:
            # This is equivalent to File.from_id(file_id), but without introducing a circular dependency
            return self.__roboto_client.get(
                f"v1/files/record/{self.association.association_id}"
            ).to_dict(json_path=["data", "association_id"])

        return None

    @property
    def default_representation(self) -> typing.Optional[RepresentationRecord]:
        return self.__record.default_representation

    @property
    def end_time(self) -> typing.Optional[int]:
        return self.__record.end_time

    @property
    def file_id(self) -> typing.Optional[str]:
        if self.association.is_file:
            return self.association.association_id

        return None

    @property
    def message_count(self) -> typing.Optional[int]:
        return self.__record.message_count

    @property
    def message_paths(self) -> collections.abc.Sequence[MessagePathRecord]:
        return self.__record.message_paths

    @property
    def metadata(self) -> dict[str, typing.Any]:
        return dict(self.__record.metadata)

    @property
    def modified(self) -> datetime:
        return self.__record.modified

    @property
    def modified_by(self) -> str:
        return self.__record.modified_by

    @property
    def name(self) -> str:
        return self.__record.topic_name

    @property
    def org_id(self) -> str:
        return self.__record.org_id

    @property
    def record(self) -> TopicRecord:
        """Topic representation in the Roboto database.

        This property is on the path to deprecation. All ``TopicRecord`` attributes
        are accessible directly using a ``Topic`` instance.
        """

        return self.__record

    @property
    def schema_checksum(self) -> typing.Optional[str]:
        return self.__record.schema_checksum

    @property
    def schema_name(self) -> typing.Optional[str]:
        return self.__record.schema_name

    @property
    def schema_version(self) -> int:
        return self.__record.schema_version

    @property
    def start_time(self) -> typing.Optional[int]:
        return self.__record.start_time

    @property
    def topic_id(self) -> str:
        return self.__record.topic_id

    @property
    def topic_name(self) -> str:
        return self.__record.topic_name

    @property
    def url_quoted_name(self) -> str:
        return urllib.parse.quote_plus(self.name)

    # TODO: should this method return MessagePath instead?
    def add_message_path(
        self,
        message_path: str,
        data_type: str,
        canonical_data_type: CanonicalDataType,
        metadata: typing.Optional[dict[str, typing.Any]] = None,
    ) -> MessagePathRecord:

        request = AddMessagePathRequest(
            message_path=message_path,
            data_type=data_type,
            canonical_data_type=canonical_data_type,
            metadata=metadata or {},
        )

        encoded_association = self.association.url_encode()
        response = self.__roboto_client.post(
            f"v1/topics/association/{encoded_association}/name/{self.url_quoted_name}/message-path",
            data=request,
            owner_org_id=self.org_id,
        )
        message_path_record = response.to_record(MessagePathRecord)
        self.refresh()
        return message_path_record

    def add_message_path_representation(
        self,
        message_path_id: str,
        association: Association,
        storage_format: RepresentationStorageFormat,
        version: int,
    ) -> RepresentationRecord:
        encoded_association = self.association.url_encode()

        request = AddMessagePathRepresentationRequest(
            association=association,
            message_path_id=message_path_id,
            storage_format=storage_format,
            version=version,
        )

        response = self.__roboto_client.post(
            f"v1/topics/association/{encoded_association}/name/{self.url_quoted_name}/message-path/representation",
            data=request,
            owner_org_id=self.org_id,
        )
        representation_record = response.to_record(RepresentationRecord)
        self.refresh()
        return representation_record

    def delete(self) -> None:
        encoded_association = self.association.url_encode()
        self.__roboto_client.delete(
            f"v1/topics/association/{encoded_association}/name/{self.url_quoted_name}",
            owner_org_id=self.org_id,
        )

    def get_data(
        self,
        message_paths_include: typing.Optional[collections.abc.Sequence[str]] = None,
        message_paths_exclude: typing.Optional[collections.abc.Sequence[str]] = None,
        start_time: typing.Optional[Time] = None,
        end_time: typing.Optional[Time] = None,
        cache_dir: typing.Union[str, pathlib.Path, None] = None,
    ) -> collections.abc.Generator[dict[str, typing.Any], None, None]:
        """
        Return this topic's underlying data.
        Each yielded datum is a dictionary that matches this topic's schema.

        If ``message_paths_include`` or ``message_paths_exclude`` are defined,
        they should be dot notation paths that match attributes of individual data records.

        If ``start_time`` or ``end_time`` are defined,
        they should either be integers that represent nanoseconds since UNIX epoch,
        or convertible to such by :py:func:`~roboto.time.to_epoch_nanoseconds`.
        Either or both may be omitted.
        ``start_time`` is inclusive, while ``end_time`` is exclusive.

        If ``cache_dir`` is defined, topic data will be downloaded to this location if necessary.
        If not provided, ``cache_dir`` defaults to
        :py:attr:`~roboto.domain.topics.topic_data_service.TopicDataService.DEFAULT_CACHE_DIR`.

        For each example below, assume the following is a sample datum record that can be found in this topic:

        ::

            {
                "angular_velocity": {
                    "x": <uint32>,
                    "y": <uint32>,
                    "z": <uint32>
                },
                "orientation": {
                    "x": <uint32>,
                    "y": <uint32>,
                    "z": <uint32>,
                    "w": <uint32>
                }
            }

        Examples:
            Print all data to stdout.

            >>> topic = Topic.from_name_and_file(...)
            >>> for record in topic.get_data():
            >>>      print(record)

            Only include the `"angular_velocity"` sub-object, but filter out its `"y"` property.

            >>> topic = Topic.from_name_and_file(...)
            >>> for record in topic.get_data(
            >>>   message_paths_include=["angular_velocity"],
            >>>   message_paths_exclude=["angular_velocity.y"],
            >>> ):
            >>>      print(record)

            Only include data between two timestamps:

            >>> topic = Topic.from_name_and_file(...)
            >>> for record in topic.get_data(
            >>>   start_time=1722870127699468923,
            >>>   end_time=1722870127699468924,
            >>> ):
            >>>      print(record)

            Collect all topic data into a dataframe. Requires installing the ``roboto[analytics]`` extra.

            >>> topic = Topic.from_name_and_file(...)
            >>> df = topic.get_data_as_df()

        """
        yield from self.__topic_data_service.get_data(
            topic_id=self.__record.topic_id,
            message_paths_include=message_paths_include,
            message_paths_exclude=message_paths_exclude,
            start_time=start_time,
            end_time=end_time,
            cache_dir_override=cache_dir,
        )

    def get_data_as_df(
        self,
        message_paths_include: typing.Optional[collections.abc.Sequence[str]] = None,
        message_paths_exclude: typing.Optional[collections.abc.Sequence[str]] = None,
        start_time: typing.Optional[Time] = None,
        end_time: typing.Optional[Time] = None,
        cache_dir: typing.Union[str, pathlib.Path, None] = None,
    ) -> pandas.DataFrame:
        """
        Return this topic's underlying data as a pandas DataFrame.

        Requires installing this package using the ``roboto[analytics]`` extra.

        See :py:meth:`~roboto.domain.topics.topic.Topic.get_data` for more information on the parameters.
        """
        pandas = import_optional_dependency("pandas", "analytics")

        df = pandas.json_normalize(
            data=list(
                self.get_data(
                    message_paths_include=message_paths_include,
                    message_paths_exclude=message_paths_exclude,
                    start_time=start_time,
                    end_time=end_time,
                    cache_dir=cache_dir,
                )
            )
        )
        return df.set_index(TopicDataService.LOG_TIME_ATTR_NAME)

    def get_message_path(self, message_path: str) -> MessagePath:
        for message_path_record in self.__record.message_paths:
            if message_path_record.message_path == message_path:
                return MessagePath(
                    message_path_record,
                    roboto_client=self.__roboto_client,
                    topic_data_service=self.__topic_data_service,
                )

        raise ValueError(
            f"Topic '{self.name}' does not have a message path matching '{message_path}'"
        )

    def refresh(self) -> None:
        topic = Topic.from_id(self.topic_id, self.__roboto_client)
        self.__record = topic.__record

    def set_default_representation(
        self,
        association: Association,
        storage_format: RepresentationStorageFormat,
        version: int,
    ) -> RepresentationRecord:
        encoded_association = self.association.url_encode()
        request = SetDefaultRepresentationRequest(
            association=association,
            storage_format=storage_format,
            version=version,
        )
        response = self.__roboto_client.post(
            f"v1/topics/association/{encoded_association}/name/{self.url_quoted_name}/representation",
            data=request,
            owner_org_id=self.org_id,
        )
        representation_record = response.to_record(RepresentationRecord)
        self.refresh()
        return representation_record

    def to_association(self) -> Association:
        return Association.topic(self.topic_id)

    def update(
        self,
        end_time: typing.Union[typing.Optional[int], NotSetType] = NotSet,
        message_count: typing.Union[int, NotSetType] = NotSet,
        schema_checksum: typing.Union[typing.Optional[str], NotSetType] = NotSet,
        schema_name: typing.Union[typing.Optional[str], NotSetType] = NotSet,
        start_time: typing.Union[typing.Optional[int], NotSetType] = NotSet,
        metadata_changeset: typing.Union[MetadataChangeset, NotSetType] = NotSet,
        message_path_changeset: typing.Union[MessagePathChangeset, NotSetType] = NotSet,
    ) -> "Topic":
        """Updates a topic's attributes and (optionally) its message paths.

        Args:
            schema_name:
              topic schema name. Setting to ``None`` clears the attribute.
            schema_checksum:
              topic schema checksum. Setting to ``None`` clears the attribute.
            start_time:
              topic data start time, in epoch nanoseconds.
              Must be non-negative. Setting to ``None`` clears the attribute.
            end_time:
              topic data end time, in epoch nanoseconds. Must be non-negative, and greater than `start_time`.
              Setting to `None` clears the attribute.
            message_count:
              number of messages recorded for this topic. Must be non-negative.
            metadata_changeset:
              a set of changes to apply to the topic's metadata
            message_path_changeset:
              a set of additions, deletions or updates to this topic's message paths.
              Updating or deleting non-existent message paths has no effect.
              Attempting to (re-)add existing message paths raises ``RobotoConflictException``,
              unless the changeset's ``replace_all`` flag is set to ``True``

        Returns:
            this ``Topic`` object with any updates applied

        Raises:
            RobotoInvalidRequestException:
              if any method argument has an invalid value, e.g. a negative ``message_count``
            RobotoConflictException:
              if, as part of the update, an attempt is made to add an already extant
              message path, and to this topic, and ``replace_all`` is not toggled
              on the ``message_path_changeset``
        """

        request = remove_not_set(
            UpdateTopicRequest(
                end_time=end_time,
                message_count=message_count,
                schema_checksum=schema_checksum,
                schema_name=schema_name,
                start_time=start_time,
                metadata_changeset=metadata_changeset,
                message_path_changeset=message_path_changeset,
            )
        )

        response = self.__roboto_client.put(
            f"v1/topics/id/{self.topic_id}",
            data=request,
            owner_org_id=self.org_id,
        )

        record = response.to_record(TopicRecord)
        self.__record = record

        if is_set(message_path_changeset) and message_path_changeset.has_changes():
            self.refresh()

        return self

    def update_message_path(
        self,
        message_path: str,
        metadata_changeset: typing.Union[TaglessMetadataChangeset, NotSetType] = NotSet,
        data_type: typing.Union[str, NotSetType] = NotSet,
        canonical_data_type: typing.Union[CanonicalDataType, NotSetType] = NotSet,
    ) -> MessagePath:
        """Updates the metadata and (optionally) other attributes of a message path.

        Args:
            message_path:
              name of the message path to update
            metadata_changeset:
              metadata changeset to apply to any existing metadata
            data_type:
              native (i.e. application-specific) message path data type.
            canonical_data_type:
              canonical Roboto data type corresponding to the native data type.

        Returns:
            A ``MessagePath`` representing the updated message path

        Raises:
            RobotoNotFoundException: if no such message path exists for this topic
        """

        request = remove_not_set(
            UpdateMessagePathRequest(
                message_path=message_path,
                metadata_changeset=metadata_changeset,
                data_type=data_type,
                canonical_data_type=canonical_data_type,
            )
        )

        encoded_association = self.association.url_encode()
        response = self.__roboto_client.put(
            f"v1/topics/association/{encoded_association}/name/{self.url_quoted_name}/message-path",
            data=request,
            owner_org_id=self.org_id,
        )

        message_path_record = response.to_record(MessagePathRecord)
        self.refresh()

        return MessagePath(
            record=message_path_record,
            roboto_client=self.__roboto_client,
            topic_data_service=self.__topic_data_service,
        )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing

from ..http import HttpClient, RobotoClient


class CLIContext:
    __roboto_service_base_url: typing.Optional[str]
    __http: typing.Optional[HttpClient]
    extensions: dict[str, typing.Any]
    roboto_client: RobotoClient

    @property
    def roboto_service_base_url(self) -> str:
        if self.__roboto_service_base_url is None:
            raise ValueError("roboto_service_base_url is unset")

        return self.__roboto_service_base_url

    @roboto_service_base_url.setter
    def roboto_service_base_url(self, roboto_service_base_url: str) -> None:
        self.__roboto_service_base_url = roboto_service_base_url

    @property
    def http_client(self) -> HttpClient:
        # Necessary since http is lazy set after parsing args
        if self.__http is None:
            raise ValueError("Unset HTTP client!")

        return self.__http

    @http_client.setter
    def http_client(self, http: HttpClient) -> None:
        self.__http = http



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

RESOURCE_OWNER_OVERRIDE_HEADER = "X-Roboto-Resource-Owner-Id"
"""Header to specify the organization that owns the resource being accessed."""

RESOURCE_OWNER_OVERRIDE_QUERY_PARAM = "robotoResourceOwnerId"
"""Query parameter to specify the organization that owns the resource being accessed."""

ORG_OVERRIDE_HEADER = "X-Roboto-Org-Id"
"""Header to specify the organization that the user is acting on behalf of."""

ORG_OVERRIDE_QUERY_PARAM = "robotoOrgId"
"""Query parameter to specify the organization that the user is acting on behalf of."""

USER_OVERRIDE_HEADER = "X-Roboto-User-Id"
"""Header to specify the user that is performing the REST operation."""

USER_OVERRIDE_QUERY_PARAM = "robotoUserId"
""""Query parameter to specify the user that is performing the REST operation."""

BEARER_TOKEN_HEADER = "X-Roboto-Bearer-Token"
"""Bearer token which is parsed as a JWT to provide additional request context for invocations"""



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json
import time

from ...domain import actions
from ..command import RobotoCommand
from ..context import CLIContext


def status(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    if not args.tail:
        invocation = actions.Invocation.from_id(
            args.invocation_id,
            roboto_client=context.roboto_client,
        )
        print(
            json.dumps(
                [
                    status_record.to_presentable_dict()
                    for status_record in invocation.status_log
                ],
                indent=2,
            )
        )
        return

    printed: set[actions.InvocationStatus] = set()
    invocation = actions.Invocation.from_id(
        args.invocation_id,
        roboto_client=context.roboto_client,
    )
    try:
        while True:
            status_records_to_print = [
                status_record
                for status_record in invocation.status_log
                if status_record.status not in printed
            ]
            if status_records_to_print:
                for status_record in status_records_to_print:
                    printed.add(status_record.status)
                    print(json.dumps(status_record.to_presentable_dict(), indent=2))

            if invocation.reached_terminal_status:
                break

            time.sleep(1)
            invocation.refresh()
    except KeyboardInterrupt:
        pass  # Swallow


def status_parser(parser: argparse.ArgumentParser):
    parser.add_argument("invocation_id")
    parser.add_argument("--tail", required=False, action="store_true")


status_command = RobotoCommand(
    name="status",
    logic=status,
    setup_parser=status_parser,
    command_kwargs={"help": "Get invocation status."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import enum
import typing
import urllib.parse

import pydantic

from .exceptions import (
    RobotoIllegalArgumentException,
)


class AssociationType(enum.Enum):
    """AssociationType is the Roboto domain entity type of the association."""

    Dataset = "dataset"
    File = "file"
    Topic = "topic"
    MessagePath = "message_path"


class Association(pydantic.BaseModel):
    """Use to declare an association between two Roboto entities."""

    URL_ENCODING_SEP: typing.ClassVar[str] = ":"

    @staticmethod
    def group_by_type(
        associations: collections.abc.Collection["Association"],
    ) -> collections.abc.Mapping[
        AssociationType, collections.abc.Sequence["Association"]
    ]:
        response: dict[AssociationType, list[Association]] = {}

        for association in associations:
            if association.association_type not in response:
                response[association.association_type] = []
            response[association.association_type].append(association)

        return response

    @classmethod
    def from_url_encoded_value(cls, encoded: str) -> "Association":
        """Reverse of Association::url_encode."""
        unquoted = urllib.parse.unquote_plus(encoded)

        association_type, association_id, *rest = unquoted.split(cls.URL_ENCODING_SEP)
        association_version: int | None = None
        if rest:
            association_version = int(rest[0])

        try:
            return cls(
                association_id=association_id,
                association_type=AssociationType(association_type),
                association_version=association_version,
            )
        except ValueError:
            raise RobotoIllegalArgumentException(
                f"Invalid association type '{association_type}'"
            ) from None

    @classmethod
    def coalesce(
        cls,
        associations: typing.Optional[collections.abc.Collection["Association"]] = None,
        dataset_ids: typing.Optional[collections.abc.Collection[str]] = None,
        file_ids: typing.Optional[collections.abc.Collection[str]] = None,
        topic_ids: typing.Optional[collections.abc.Collection[str]] = None,
        message_path_ids: typing.Optional[collections.abc.Collection[str]] = None,
        throw_on_empty: bool = False,
    ) -> list["Association"]:
        coalesced: list[Association] = []

        if associations:
            coalesced.extend(associations)

        if dataset_ids:
            coalesced.extend(cls.dataset(dataset_id) for dataset_id in dataset_ids)

        if file_ids:
            coalesced.extend(cls.file(file_id) for file_id in file_ids)

        if message_path_ids:
            coalesced.extend(cls.msgpath(msgpath_id) for msgpath_id in message_path_ids)

        if topic_ids:
            coalesced.extend(cls.topic(topic_id) for topic_id in topic_ids)

        if len(coalesced) == 0 and throw_on_empty:
            raise RobotoIllegalArgumentException(
                "At least one association must be provided"
            )

        return coalesced

    @classmethod
    def dataset(cls, dataset_id: str) -> "Association":
        return cls(association_id=dataset_id, association_type=AssociationType.Dataset)

    @classmethod
    def file(cls, file_id: str, version: typing.Optional[int] = None):
        return cls(
            association_id=file_id,
            association_type=AssociationType.File,
            association_version=version,
        )

    @classmethod
    def topic(cls, topic_id: str):
        return cls(association_id=topic_id, association_type=AssociationType.Topic)

    @classmethod
    def msgpath(cls, msgpath_id: str) -> "Association":
        return cls(
            association_id=msgpath_id, association_type=AssociationType.MessagePath
        )

    association_id: str
    """Roboto identifier"""

    association_type: AssociationType
    """association_type is the Roboto domain entity type of the association."""

    association_version: int | None = None
    """association_version is the Roboto domain entity version of the association, if it exists."""

    @property
    def is_dataset(self) -> bool:
        return self.association_type == AssociationType.Dataset

    @property
    def is_file(self) -> bool:
        return self.association_type == AssociationType.File

    @property
    def is_msgpath(self) -> bool:
        return self.association_type == AssociationType.MessagePath

    @property
    def is_topic(self) -> bool:
        return self.association_type == AssociationType.Topic

    def url_encode(self) -> str:
        """Association encoded in a URL path segment ready format."""

        parts = [
            self.association_type.value,
            self.association_id,
        ]

        if self.association_version is not None:
            parts.append(str(self.association_version))

        return urllib.parse.quote_plus(Association.URL_ENCODING_SEP.join(parts))



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import errno
import http
import http.client
import logging
import typing
import urllib.error
import urllib.parse
import urllib.request
import urllib.response

import tenacity
import tenacity.wait

from roboto.exceptions import (
    ClientError,
    HttpError,
    ServerError,
)

from ..logging import LOGGER_NAME
from .request import (
    HttpRequest,
    HttpRequestDecorator,
    RetryWaitFn,
)
from .requester import (
    ROBOTO_REQUESTER_HEADER,
    RobotoRequester,
)
from .response import HttpResponse

logger = logging.getLogger(LOGGER_NAME)

DEFAULT_HEADERS = object()


class is_expected_to_be_transient:
    """Retry predicate that returns True if the exception is expected to be transient."""

    __request: HttpRequest

    def __init__(self, request: HttpRequest):
        self.__request = request

    def __call__(self, exc: typing.Any) -> bool:
        if isinstance(exc, http.client.RemoteDisconnected):
            logger.warning("Remote host closed connection", exc_info=exc)
            if not self.__request_expected_to_be_idempotent():
                # Request may have be processed, we don't know.
                return False

            return True

        if isinstance(exc, urllib.error.HTTPError):
            try:
                status_code = http.HTTPStatus(int(exc.code))

                if not self.__request_expected_to_be_idempotent():
                    # 500 and 504 are not safe to retry if the request is not idempotent
                    if status_code in (
                        http.HTTPStatus.REQUEST_TIMEOUT,
                        http.HTTPStatus.INTERNAL_SERVER_ERROR,
                        http.HTTPStatus.GATEWAY_TIMEOUT,
                    ):
                        return False

                return status_code in (
                    http.HTTPStatus.REQUEST_TIMEOUT,
                    http.HTTPStatus.INTERNAL_SERVER_ERROR,
                    http.HTTPStatus.TOO_MANY_REQUESTS,
                    http.HTTPStatus.BAD_GATEWAY,
                    http.HTTPStatus.SERVICE_UNAVAILABLE,
                    http.HTTPStatus.GATEWAY_TIMEOUT,
                )
            except ValueError:
                return False

        # Must come after the isinstance check of HTTPError,
        # which is a subclass of URLError
        if isinstance(exc, urllib.error.URLError):
            # Connection reset by peer, probably transient
            conn_reset_error = (
                isinstance(exc.reason, OSError) and exc.reason.errno == errno.ECONNRESET
            ) or "Connection reset by peer" in str(exc.reason)
            if conn_reset_error and self.__request_expected_to_be_idempotent():
                return True

            # DNS error, probably transient
            return "Temporary failure in name resolution" in str(exc.reason)

        return False

    def __request_expected_to_be_idempotent(self) -> bool:
        return self.__request.idempotent is True


class HttpClient:
    __base_headers: dict[str, str]
    __default_auth: typing.Optional[HttpRequestDecorator]
    __default_endpoint: typing.Optional[str]

    def __init__(
        self,
        base_headers: typing.Optional[dict[str, str]] = None,
        default_endpoint: typing.Optional[str] = None,
        default_auth: typing.Optional[HttpRequestDecorator] = None,
        requester: typing.Optional[RobotoRequester] = None,
    ):
        self.__base_headers = base_headers if base_headers is not None else {}

        if requester is not None:
            self.set_requester(requester)

        self.__default_auth = default_auth
        self.__default_endpoint = default_endpoint

    def delete(
        self,
        url: str,
        data: typing.Any = None,
        headers: typing.Optional[dict] = None,
        idempotent: bool = True,
        retry_wait: typing.Optional[RetryWaitFn] = None,
    ) -> HttpResponse:
        request = HttpRequest(
            url=url,
            method="DELETE",
            data=data,
            headers=headers,
            idempotent=idempotent,
            retry_wait=retry_wait,
        )
        return self.__request(request)

    def get(
        self,
        url: str,
        headers: typing.Optional[dict] = None,
        retry_wait: typing.Optional[RetryWaitFn] = None,
        idempotent: bool = True,
    ) -> HttpResponse:
        request = HttpRequest(
            url=url,
            method="GET",
            headers=headers,
            retry_wait=retry_wait,
            idempotent=idempotent,
        )
        return self.__request(request)

    def post(
        self,
        url,
        data: typing.Any = None,
        headers: typing.Optional[dict] = None,
        idempotent: bool = False,
        retry_wait: typing.Optional[RetryWaitFn] = None,
    ) -> HttpResponse:
        request = HttpRequest(
            url=url,
            method="POST",
            data=data,
            headers=headers,
            idempotent=idempotent,
            retry_wait=retry_wait,
        )
        return self.__request(request)

    def patch(
        self,
        url,
        data: typing.Any = None,
        headers: typing.Optional[dict] = None,
        idempotent: bool = True,
        retry_wait: typing.Optional[RetryWaitFn] = None,
    ) -> HttpResponse:
        request = HttpRequest(
            url=url,
            method="PATCH",
            data=data,
            headers=headers,
            idempotent=idempotent,
            retry_wait=retry_wait,
        )
        return self.__request(request)

    def put(
        self,
        url,
        data: typing.Any = None,
        headers: typing.Optional[dict] = None,
        idempotent: bool = True,
        retry_wait: typing.Optional[RetryWaitFn] = None,
    ) -> HttpResponse:
        request = HttpRequest(
            url=url,
            method="PUT",
            data=data,
            headers=headers,
            idempotent=idempotent,
            retry_wait=retry_wait,
        )
        return self.__request(request)

    def set_requester(self, requester: RobotoRequester):
        self.__base_headers[ROBOTO_REQUESTER_HEADER] = requester.model_dump_json(
            exclude_none=True
        )

    def url(self, path: str) -> str:
        if self.__default_endpoint is None:
            raise ValueError(
                "HttpClient.url called for client with no default endpoint."
            )
        return f"{self.__default_endpoint}/{path}"

    @property
    def auth_decorator(self) -> typing.Optional[HttpRequestDecorator]:
        return self.__default_auth

    def __request_headers(self, request_ctx: HttpRequest) -> dict[str, str]:
        headers = self.__base_headers.copy()

        if request_ctx.headers:
            headers.update(request_ctx.headers)

        return headers

    def __request(self, request_ctx: HttpRequest) -> HttpResponse:
        if self.__default_auth is not None:
            request_ctx = self.__default_auth(request_ctx)

        logger.debug("%s", request_ctx)

        headers = self.__request_headers(request_ctx)

        try:
            for attempt in tenacity.Retrying(
                retry=tenacity.retry_if_exception(
                    is_expected_to_be_transient(request_ctx)
                ),
                stop=tenacity.stop_after_attempt(10),
                reraise=True,
                wait=self._wait(request_ctx.retry_wait),
            ):
                with attempt:
                    request = urllib.request.Request(
                        request_ctx.url, method=request_ctx.method
                    )

                    req_body = request_ctx.body
                    if req_body is not None:
                        request.data = req_body

                    for key, value in headers.items():
                        request.add_header(key, value)

                    response = HttpResponse(urllib.request.urlopen(request))
                    logger.debug("Response: %s %s", response.status, response.headers)
                    return response
        except urllib.error.HTTPError as exc:
            logger.debug("HTTPError: %s", exc, exc_info=True)
            status_code = exc.code
            if status_code > 399 and status_code < 500:
                raise ClientError(exc) from None
            elif status_code > 499:
                raise ServerError(exc) from None
            else:
                raise HttpError(exc) from None
        except urllib.error.URLError as exc:
            logger.debug("URLError: %s", exc, exc_info=True)
            if (
                isinstance(exc.reason, OSError)
                and exc.reason.errno == errno.ECONNREFUSED
            ):
                raise ConnectionRefusedError(
                    f"Couldn't connect to endpoint {request_ctx.url}"
                ) from None
            else:
                raise
        except Exception:
            logger.debug("Catchall Exception", exc_info=True)
            raise

        raise RuntimeError("Unreachable")

    def _wait(self, waiter: RetryWaitFn) -> tenacity.wait.wait_base:
        class Waiter(tenacity.wait.wait_base):
            def __call__(self, retry_state: tenacity.RetryCallState) -> float:
                return waiter(retry_state, None) / 1000

        return Waiter()



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import datetime
import typing

from ...exceptions import (
    RobotoIllegalArgumentException,
)
from ...http import RobotoClient
from ...query import QuerySpecification
from ...sentinels import (
    NotSet,
    NotSetType,
    is_set,
    remove_not_set,
)
from ...updates import MetadataChangeset
from .action_operations import (
    CreateActionRequest,
    SetActionAccessibilityRequest,
    UpdateActionRequest,
)
from .action_record import (
    Accessibility,
    ActionParameter,
    ActionParameterChangeset,
    ActionRecord,
    ActionReference,
    ComputeRequirements,
    ContainerParameters,
)
from .invocation import Invocation
from .invocation_operations import (
    CreateInvocationRequest,
)
from .invocation_record import (
    InvocationDataSourceType,
    InvocationRecord,
    InvocationSource,
)

SHORT_DESCRIPTION_LENGTH = 140


class Action:
    """
    A reusable function to process, transform or analyze data in Roboto.
    """

    __record: ActionRecord
    __roboto_client: RobotoClient

    @classmethod
    def create(
        cls,
        name: str,
        compute_requirements: typing.Optional[ComputeRequirements] = None,
        container_parameters: typing.Optional[ContainerParameters] = None,
        description: typing.Optional[str] = None,
        inherits: typing.Optional[ActionReference] = None,
        metadata: typing.Optional[dict[str, typing.Any]] = None,
        parameters: typing.Optional[list[ActionParameter]] = None,
        short_description: typing.Optional[str] = None,
        tags: typing.Optional[list[str]] = None,
        timeout: typing.Optional[int] = None,
        uri: typing.Optional[str] = None,
        caller_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Action":
        request = CreateActionRequest(
            compute_requirements=compute_requirements,
            container_parameters=container_parameters,
            description=description,
            inherits=inherits,
            metadata=metadata or {},
            name=name,
            parameters=parameters or [],
            short_description=short_description,
            tags=tags or [],
            timeout=timeout,
            uri=uri,
        )

        if (
            is_set(request.short_description)
            and request.short_description is not None
            and len(request.short_description) > SHORT_DESCRIPTION_LENGTH
        ):
            raise RobotoIllegalArgumentException(
                f"Short description exceeds max length of {SHORT_DESCRIPTION_LENGTH} characters"
            )

        roboto_client = RobotoClient.defaulted(roboto_client)

        response = roboto_client.post(
            "v1/actions",
            data=request,
            caller_org_id=caller_org_id,
        )
        record = response.to_record(ActionRecord)
        return cls(record, roboto_client)

    @classmethod
    def from_name(
        cls,
        name: str,
        digest: typing.Optional[str] = None,
        owner_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Action":
        roboto_client = RobotoClient.defaulted(roboto_client)
        query_params = {"digest": digest} if digest else None
        response = roboto_client.get(
            f"v1/actions/{name}",
            query=query_params,
            owner_org_id=owner_org_id,
        )
        record = response.to_record(ActionRecord)
        return cls(record, roboto_client)

    @classmethod
    def query(
        cls,
        spec: typing.Optional[QuerySpecification] = None,
        accessibility: Accessibility = Accessibility.Organization,
        owner_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> collections.abc.Generator["Action", None, None]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        spec = spec or QuerySpecification()

        known = set(ActionRecord.model_fields.keys())
        actual = set()
        for field in spec.fields():
            # Support dot notation for nested fields
            # E.g., "metadata.SoftwareVersion"
            if "." in field:
                actual.add(field.split(".")[0])
            else:
                actual.add(field)
        unknown = actual - known
        if unknown:
            plural = len(unknown) > 1
            msg = (
                "are not known attributes of Action"
                if plural
                else "is not a known attribute of Action"
            )
            raise ValueError(f"{unknown} {msg}. Known attributes: {known}")

        url_path = (
            "v1/actions/query"
            if accessibility == Accessibility.Organization
            else "v1/actions/query/actionhub"
        )

        while True:
            response = roboto_client.post(
                url_path,
                data=spec,
                idempotent=True,
                owner_org_id=owner_org_id,
            )
            paginated_results = response.to_paginated_list(ActionRecord)
            for record in paginated_results.items:
                yield cls(record, roboto_client)
            if paginated_results.next_token:
                spec.after = paginated_results.next_token
            else:
                break

    def __init__(
        self,
        record: ActionRecord,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> None:
        self.__record = record
        self.__roboto_client = RobotoClient.defaulted(roboto_client)

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def accessibility(self) -> Accessibility:
        return self.__record.accessibility

    @property
    def compute_requirements(self) -> typing.Optional[ComputeRequirements]:
        return self.__record.compute_requirements

    @property
    def container_parameters(self) -> typing.Optional[ContainerParameters]:
        return self.__record.container_parameters

    @property
    def created(self) -> datetime.datetime:
        return self.__record.created

    @property
    def created_by(self) -> str:
        return self.__record.created_by

    @property
    def digest(self) -> str:
        return (
            self.__record.digest
            if self.__record.digest
            else self.__record.compute_digest()
        )

    @property
    def inherits_from(self) -> typing.Optional[ActionReference]:
        return self.__record.inherits

    @property
    def modified(self) -> datetime.datetime:
        return self.__record.modified

    @property
    def modified_by(self) -> str:
        return self.__record.modified_by

    @property
    def name(self) -> str:
        return self.__record.name

    @property
    def org_id(self) -> str:
        return self.__record.org_id

    @property
    def parameters(self) -> collections.abc.Sequence[ActionParameter]:
        return [param.model_copy() for param in self.__record.parameters]

    @property
    def record(self) -> ActionRecord:
        return self.__record

    @property
    def uri(self) -> typing.Optional[str]:
        return self.__record.uri

    @property
    def timeout(self) -> typing.Optional[int]:
        return self.__record.timeout

    def delete(self) -> None:
        self.__roboto_client.delete(
            f"v1/actions/{self.name}",
            owner_org_id=self.org_id,
        )

    def invoke(
        self,
        data_source_id: str,
        data_source_type: InvocationDataSourceType,
        input_data: list[str],
        invocation_source: InvocationSource,
        compute_requirement_overrides: typing.Optional[ComputeRequirements] = None,
        container_parameter_overrides: typing.Optional[ContainerParameters] = None,
        idempotency_id: typing.Optional[str] = None,
        invocation_source_id: typing.Optional[str] = None,
        parameter_values: typing.Optional[dict[str, typing.Any]] = None,
        timeout: typing.Optional[int] = None,
        caller_org_id: typing.Optional[str] = None,
    ) -> Invocation:
        request = CreateInvocationRequest(
            data_source_id=data_source_id,
            data_source_type=data_source_type,
            input_data=input_data,
            invocation_source=invocation_source,
            compute_requirement_overrides=compute_requirement_overrides,
            container_parameter_overrides=container_parameter_overrides,
            idempotency_id=idempotency_id,
            invocation_source_id=invocation_source_id,
            parameter_values=parameter_values,
            timeout=timeout,
        )

        if request.timeout is None:
            request.timeout = self.__record.timeout

        response = self.__roboto_client.post(
            f"v1/actions/{self.name}/invoke",
            data=request,
            # invocation record will be created in the effective org of the requestor
            caller_org_id=caller_org_id,
            # owner_org_id is the org that owns the action to be invoked
            owner_org_id=self.org_id,
            query={"digest": self.digest},
        )
        record = response.to_record(InvocationRecord)
        return Invocation(record, self.__roboto_client)

    def set_accessibility(self, accessibility: Accessibility) -> "Action":
        request = SetActionAccessibilityRequest(accessibility=accessibility)
        response = self.__roboto_client.put(
            f"v1/actions/{self.name}/accessibility",
            data=request,
            owner_org_id=self.org_id,
        )
        record = response.to_record(ActionRecord)
        self.__record = record
        return self

    def to_dict(self) -> dict[str, typing.Any]:
        return self.__record.model_dump(mode="json")

    def update(
        self,
        compute_requirements: typing.Optional[
            typing.Union[ComputeRequirements, NotSetType]
        ] = NotSet,
        container_parameters: typing.Optional[
            typing.Union[ContainerParameters, NotSetType]
        ] = NotSet,
        description: typing.Optional[typing.Union[str, NotSetType]] = NotSet,
        inherits: typing.Optional[typing.Union[ActionReference, NotSetType]] = NotSet,
        metadata_changeset: typing.Union[MetadataChangeset, NotSetType] = NotSet,
        parameter_changeset: typing.Union[
            ActionParameterChangeset, NotSetType
        ] = NotSet,
        short_description: typing.Optional[typing.Union[str, NotSetType]] = NotSet,
        timeout: typing.Optional[typing.Union[int, NotSetType]] = NotSet,
        uri: typing.Optional[typing.Union[str, NotSetType]] = NotSet,
    ) -> "Action":
        request = remove_not_set(
            UpdateActionRequest(
                compute_requirements=compute_requirements,
                container_parameters=container_parameters,
                description=description,
                inherits=inherits,
                metadata_changeset=metadata_changeset,
                parameter_changeset=parameter_changeset,
                uri=uri,
                short_description=short_description,
                timeout=timeout,
            )
        )

        if (
            is_set(request.short_description)
            and request.short_description is not None
            and len(request.short_description) > SHORT_DESCRIPTION_LENGTH
        ):
            raise RobotoIllegalArgumentException(
                f"Short description exceeds max length of {SHORT_DESCRIPTION_LENGTH} characters"
            )

        response = self.__roboto_client.put(
            f"v1/actions/{self.name}",
            data=request,
            owner_org_id=self.org_id,
        )
        record = response.to_record(ActionRecord)
        self.__record = record
        return self
